#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
# pyvirtualbench library: Copyright (c) 2015 Charles Armstrap <charles@armstrap.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Requires 'pyvirtualbench' package:
# https://github.com/armstrap/armstrap-pyvirtualbench

import logging
import re
# ctypes only required for VirtualBench_Direct class
from ctypes import (c_bool, c_size_t, c_double, c_uint8, c_int32, c_uint32,
                    c_int64, c_uint64, c_wchar, c_wchar_p, Structure, c_int,
                    cdll, byref)
from datetime import datetime, timezone, timedelta
import numpy as np
import pandas as pd

from pymeasure.instruments import Instrument, RangeException
from pymeasure.instruments.validators import (
    strict_discrete_set, strict_discrete_range,
    truncated_discrete_set, strict_range
)

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

try:
    # Requires 'pyvirtualbench' package:
    # https://github.com/armstrap/armstrap-pyvirtualbench
    import pyvirtualbench as pyvb
except ModuleNotFoundError as err:
    log.info('Failed loading the pyvirtualbench package. '
             + 'Check the NI VirtualBench documentation on how to '
             + 'install this external dependency. '
             + 'ImportError: {}'.format(err))
    raise


class VirtualBench_Direct(pyvb.PyVirtualBench):
    """ Represents National Instruments Virtual Bench main frame.
    This class provides direct access to the armstrap/pyvirtualbench
    Python wrapper.
    """

    def __init__(self, device_name='', name='VirtualBench'):
        ''' Initialize the VirtualBench library.  This must be called at least
            once for the application. The 'version' parameter must be set to
            the NIVB_LIBRARY_VERSION constant.
        '''
        self.device_name = device_name
        self.name = name
        self.nilcicapi = cdll.LoadLibrary("nilcicapi")
        self.library_handle = c_int(0)
        status = self.nilcicapi.niVB_Initialize(pyvb.NIVB_LIBRARY_VERSION,
                                                byref(self.library_handle))
        if (status != pyvb.Status.SUCCESS):
            raise pyvb.PyVirtualBenchException(status, self.nilcicapi,
                                               self.library_handle)
        log.info("Initializing %s." % self.name)

    def __del__(self):
        """ Ensures the connection is closed upon deletion
        """
        self.release()


class VirtualBench():
    """ Represents National Instruments Virtual Bench main frame.

    Subclasses implement the functionalities of the different modules:

        - Mixed-Signal-Oscilloscope (MSO)
        - Digital Input Output (DIO)
        - Function Generator (FGEN)
        - Power Supply (PS)
        - Serial Peripheral Interface (SPI) -> not implemented
          for pymeasure yet
        - Inter Integrated Circuit (I2C) -> not implemented for pymeasure yet

    For every module exist methods to save/load the configuration to file.
    These methods are not wrapped so far, checkout the pyvirtualbench file.

    All calibration methods and classes are not wrapped so far, since these
    are not required on a very regular basis. Also the connections via network
    are not yet implemented.
    Check the pyvirtualbench file, if you need the functionality.

    :param str device_name: Full unique device name
    :param str name: Name for display in pymeasure
    """

    def __init__(self, device_name='', name='VirtualBench'):
        ''' Initialize the VirtualBench library.  This must be called at least
        once for the application. The 'version' parameter must be set to the
        NIVB_LIBRARY_VERSION constant.
        '''
        self.device_name = device_name
        self.name = name
        self.vb = pyvb.PyVirtualBench(self.device_name)
        log.info("Initializing %s." % self.name)

    def __del__(self):
        """ Ensures the connection is closed upon deletion
        """
        if self.vb.library_handle is not None:
            self.vb.release()

    def shutdown(self):
        ''' Finalize the VirtualBench library.
        '''
        log.info("Shutting down %s" % self.name)
        self.vb.release()
        self.isShutdown = True

    def get_library_version(self):
        ''' Return the version of the VirtualBench runtime library
        '''
        return self.vb.get_library_version()

    def convert_timestamp_to_values(self, timestamp):
        """ Converts a timestamp to seconds and fractional seconds

        :param timestamp: VirtualBench timestamp
        :type timestamp: pyvb.Timestamp
        :return: (seconds_since_1970, fractional seconds)
        :rtype: (int, float)
        """
        if not isinstance(timestamp, pyvb.Timestamp):
            raise ValueError("{0} is not a VirtualBench Timestamp object"
                             .format(timestamp))
        return self.vb.convert_timestamp_to_values(timestamp)

    def convert_values_to_timestamp(self, seconds_since_1970,
                                    fractional_seconds):
        """ Converts seconds and fractional seconds to a timestamp

        :param seconds_since_1970: Date/Time in seconds since 1970
        :type seconds_since_1970: int
        :param fractional_seconds: Fractional seconds
        :type fractional_seconds: float
        :return: VirtualBench timestamp
        :rtype: pyvb.Timestamp
        """
        return self.vb.convert_values_to_timestamp(seconds_since_1970,
                                                   fractional_seconds)

    def convert_values_to_datetime(self, timestamp):
        """ Converts timestamp to datetime object

        :param timestamp: VirtualBench timestamp
        :type timestamp: pyvb.Timestamp
        :return: Timestamp as DateTime object
        :rtype: DateTime
        """
        (seconds_since_1970,
         fractional_seconds) = self.convert_timestamp_to_values(timestamp)
        fractional_seconds = timedelta(seconds=fractional_seconds)
        return (datetime.fromtimestamp(seconds_since_1970, timezone.utc) +
                fractional_seconds)

    def collapse_channel_string(self, names_in):
        """ Collapses a channel string into a comma and colon-delimited
        equivalent. Last element is the number of channels.

        :param names_in: Channel string
        :type names_in: str
        :return: Channel string with colon notation where possible,
                 number of channels
        :rtype: (str, int)
        """
        if not isinstance(names_in, str):
            raise ValueError("{0} is not a string".format(names_in))
        return self.vb.collapse_channel_string(names_in)

    def expand_channel_string(self, names_in):
        """ Expands a channel string into a comma-delimited (no colon)
        equivalent. Last element is the number of channels.
        ``'dig/0:2'`` -> ``('dig/0, dig/1, dig/2',3)``

        :param names_in: Channel string
        :type names_in: str
        :return: Channel string with all channels separated by comma,
                 number of channels
        :rtype: (str, int)
        """
        return self.vb.expand_channel_string(names_in)

    def get_calibration_information(self):
        """ Returns calibration information for the specified device,
        including the last calibration date and calibration interval.

        :return: Calibration date, recommended calibration interval in months,
                 calibration interval in months
        :rtype: (pyvb.Timestamp, int, int)
        """
        return self.vb.get_calibration_information(self.device_name)

    def acquire_digital_input_output(self, lines, reset=False):
        """ Establishes communication with the DIO module. This method should
        be called once per session.

        :param lines: Lines to acquire, reading is possible on all lines
        :type lines: str
        :param reset: Reset DIO module, defaults to False
        :type reset: bool, optional
        """
        reset = strict_discrete_set(reset, [True, False])
        self.dio = self.DigitalInputOutput(self.vb, lines, reset,
                                           vb_name=self.name)

    def acquire_power_supply(self, reset=False):
        """ Establishes communication with the PS module. This method should be
        called once per session.

        :param reset: Reset the PS module, defaults to False
        :type reset: bool, optional
        """
        reset = strict_discrete_set(reset, [True, False])
        self.ps = self.PowerSupply(self.vb, reset, vb_name=self.name)

    def acquire_function_generator(self, reset=False):
        """ Establishes communication with the FGEN module. This method should
        be called once per session.

        :param reset: Reset the FGEN module, defaults to False
        :type reset: bool, optional
        """
        reset = strict_discrete_set(reset, [True, False])
        self.fgen = self.FunctionGenerator(self.vb, reset, vb_name=self.name)

    def acquire_mixed_signal_oscilloscope(self, reset=False):
        """ Establishes communication with the MSO module. This method should
        be called once per session.

        :param reset: Reset the MSO module, defaults to False
        :type reset: bool, optional
        """
        reset = strict_discrete_set(reset, [True, False])
        self.mso = self.MixedSignalOscilloscope(self.vb, reset,
                                                vb_name=self.name)

    def acquire_digital_multimeter(self, reset=False):
        """ Establishes communication with the DMM module. This method should
        be called once per session.

        :param reset: Reset the DMM module, defaults to False
        :type reset: bool, optional
        """
        reset = strict_discrete_set(reset, [True, False])
        self.dmm = self.DigitalMultimeter(self.vb, reset=reset,
                                          vb_name=self.name)

    class VirtualBenchInstrument():
        def __init__(self, acquire_instr, reset,
                     instr_identifier, vb_name=''):
            """Initialize instrument of VirtualBench device.
            Sets class variables and provides basic methods
            common to all VB instruments.

            :param acquire_instr: Method to acquire the instrument
            :type acquire_instr: method
            :param reset: Resets the instrument
            :type reset: bool
            :param instr_identifier: Shorthand identifier, e.g. mso or fgen
            :type instr_identifier: str
            :param vb_name: Name of VB device for logging, defaults to ''
            :type vb_name: str, optional
            """
            # Parameters & Handle of VirtualBench Instance
            self._vb_handle = acquire_instr.__self__
            self._device_name = self._vb_handle.device_name
            self.name = (vb_name + " " + instr_identifier.upper()).strip()
            log.info("Initializing %s." % self.name)
            self._instrument_handle = acquire_instr(self._device_name, reset)
            self.isShutdown = False

        def __del__(self):
            """ Ensures the connection is closed upon deletion
            """
            if self.isShutdown is not True:
                self._instrument_handle.release()

        def shutdown(self):
            ''' Removes the session and deallocates any resources acquired
            during the session. If output is enabled on any channels, they
            remain in their current state.
            '''
            log.info("Shutting down %s" % self.name)
            self._instrument_handle.release()
            self.isShutdown = True

    class DigitalInputOutput(VirtualBenchInstrument):
        """ Represents Digital Input Output (DIO) Module of Virtual Bench
        device. Allows to read/write digital channels and/or set channels
        to export the start signal of FGEN module or trigger of MSO module.
        """
        def __init__(self, virtualbench, lines, reset, vb_name=''):
            """ Acquire DIO module

            :param virtualbench: VirtualBench Instance
            :type virtualbench: VirtualBench
            :param lines: Lines to acquire
            :type lines: str
            :param reset: Rest DIO module
            :type reset: bool
            """
            # Parameters & Handle of VirtualBench Instance
            self._device_name = virtualbench.device_name
            self._vb_handle = virtualbench
            self.name = vb_name + " DIO"
            # Validate lines argument
            # store line names & numbers for future reference
            (self._line_names, self._line_numbers) = self.validate_lines(
                lines, return_single_lines=True, validate_init=False)
            # Create DIO Instance
            log.info("Initializing %s." % self.name)
            self.dio = self._vb_handle.acquire_digital_input_output(
                self._line_names, reset)
            # for methods provided by super class
            self._instrument_handle = self.dio
            self.isShutdown = False

        def validate_lines(self, lines, return_single_lines=False,
                           validate_init=False):
            """Validate lines string
            Allowed patterns (case sensitive):

                - ``'VBxxxx-xxxxxxx/dig/0:7'``
                - ``'VBxxxx-xxxxxxx/dig/0'``
                - ``'dig/0'``
                - ``'VBxxxx-xxxxxxx/trig'``
                - ``'trig'``

                Allowed Line Numbers: 0-7 or trig

            :param lines: Line string to test
            :type lines: str
            :param return_single_lines: Return list of line numbers as well,
                                        defaults to False
            :type return_single_lines: bool, optional
            :param validate_init: Check if lines are initialized (in
                                  :code:`self._line_numbers`),
                                  defaults to False
            :type validate_init: bool, optional
            :return: Line string, optional list of single line numbers
            :rtype: str, optional (str, list)
            """

            def error(lines=lines):
                raise ValueError(
                    "Line specification {0} is not valid!".format(lines))

            lines = self._vb_handle.expand_channel_string(lines)[0]
            lines = lines.split(', ')
            return_lines = []
            single_lines = []
            for line in lines:
                if line == 'trig':
                    device = self._device_name
                # otherwise (device_name/)dig/line or device_name/trig
                else:
                    # split off line number by last '/'
                    try:
                        (device, line) = re.match(
                            r'(.*)(?:/)(.+)',
                            line).groups()
                    except IndexError:
                        error()
                if (line == 'trig') and (device == self._device_name):
                    single_lines.append('trig')
                    return_lines.append(self._device_name + '/' + line)
                elif int(line) in range(0, 8):
                    line = int(line)
                    single_lines.append(line)
                    # validate device name: either 'dig' or 'device_name/dig'
                    if device == 'dig':
                        pass
                    else:
                        try:
                            device = re.match(
                                r'(VB[0-9]{4}-[0-9a-zA-Z]{7})(?:/dig)',
                                device).groups()[0]
                        except (IndexError, KeyError):
                            error()
                        # device_name has to match
                        if not device == self._device_name:
                            error()
                    # constructing line references for output
                    return_lines.append((self._device_name + '/dig/%d') % line)
                else:
                    error()
                # check if lines are initialized
                if validate_init is True:
                    if line not in self._line_numbers:
                        raise ValueError(
                            "Digital Line {} is not initialized".format(line))

            # create comma separated channel string
            return_lines = ', '.join(return_lines)
            # collapse string if possible
            return_lines = self._vb_handle.collapse_channel_string(
                return_lines)[0]  # drop number of lines
            if return_single_lines is True:
                return return_lines, single_lines
            else:
                return return_lines

        def tristate_lines(self, lines):
            ''' Sets all specified lines to a high-impedance state. (Default)
            '''
            lines = self.validate_lines(lines, validate_init=True)
            self.dio.tristate_lines(lines)

        def export_signal(self, line, digitalSignalSource):
            """ Exports a signal to the specified line.

            :param line: Line string
            :type line: str
            :param digitalSignalSource: ``0`` for FGEN start or ``1``
                                        for MSO trigger
            :type digitalSignalSource: int
            """
            line = self.validate_lines(line, validate_init=True)
            digitalSignalSource_values = {"FGEN START": 0, "MSO TRIGGER": 1}
            digitalSignalSource = strict_discrete_set(
                digitalSignalSource.upper(), digitalSignalSource_values)
            digitalSignalSource = digitalSignalSource_values[
                digitalSignalSource.upper()]
            self.dio.export_signal(line, digitalSignalSource)

        def query_line_configuration(self):
            ''' Indicates the current line configurations. Tristate Lines,
            Static Lines, and Export Lines contain comma-separated
            range_data and/or colon-delimited lists of all acquired lines
            '''
            return self.dio.query_line_configuration()

        def query_export_signal(self, line):
            """ Indicates the signal being exported on the specified line.

            :param line: Line string
            :type line: str
            :return: Exported signal (FGEN start or MSO trigger)
            :rtype: enum
            """
            line = self.validate_lines(line, validate_init=True)
            return self.dio.query_export_signal(line)

        def write(self, lines, data):
            """ Writes data to the specified lines.

            :param lines: Line string
            :type lines: str
            :param data: List of data, (``True`` = High, ``False`` = Low)
            :type data: list or tuple
            """
            lines = self.validate_lines(lines, validate_init=True)
            try:
                for value in data:
                    strict_discrete_set(value, [True, False])
            except Exception:
                raise ValueError(
                    "Data {} is not iterable (list or tuple).".format(data))
            log.debug("{}: {} output {}.".format(self.name, lines, data))
            self.dio.write(lines, data)

        def read(self, lines):
            """ Reads the current state of the specified lines.

            :param lines: Line string,  requires full name specification e.g.
                          ``'VB8012-xxxxxxx/dig/0:7'`` since instrument_handle
                          is not required (only library_handle)
            :type lines: str
            :return: List of line states (HIGH/LOW)
            :rtype: list
            """
            lines = self.validate_lines(lines, validate_init=False)
            # init not necessary for readout
            return self.dio.read(lines)

        def reset_instrument(self):
            ''' Resets the session configuration to default values, and
            resets the device and driver software to a known state.
            '''
            self.dio.reset_instrument()

    class DigitalMultimeter(VirtualBenchInstrument):
        """ Represents Digital Multimeter (DMM) Module of Virtual Bench
        device. Allows to measure either DC/AC voltage or current,
        Resistance or Diodes.
        """
        def __init__(self, virtualbench, reset, vb_name=''):
            """ Acquire DMM module

            :param virtualbench: Instance of the VirtualBench class
            :type virtualbench: VirtualBench
            :param reset: Resets the instrument
            :type reset: bool
            """
            super().__init__(
                virtualbench.acquire_digital_multimeter,
                reset, 'dmm', vb_name)
            self.dmm = self._instrument_handle

        @staticmethod
        def validate_range(dmm_function, range):
            """ Checks if ``range`` is valid for the chosen ``dmm_function``

            :param int dmm_function: DMM Function
            :param range: Range value, e.g. maximum value to measure
            :type range: int or float
            :return: Range value to pass to instrument
            :rtype: int
            """
            ref_ranges = {
                0: [0.1, 1, 10, 100, 300],
                1: [0.1, 1, 10, 100, 265],
                2: [0.01, 0.1, 1, 10],
                3: [0.005, 0.05, 0.5, 5],
                4: [100, 1000, 10000, 100000, 1000000,
                    10000000, 100000000],
                }
            range = truncated_discrete_set(range, ref_ranges[dmm_function])
            return range

        def validate_dmm_function(self, dmm_function):
            """ Check if DMM function *dmm_function* exists

            :param dmm_function: DMM function index or name:

                - ``'DC_VOLTS'``, ``'AC_VOLTS'``
                - ``'DC_CURRENT'``, ``'AC_CURRENT'``
                - ``'RESISTANCE'``
                - ``'DIODE'``

            :type dmm_function: int or str
            :return: DMM function index to pass to the instrument
            :rtype: int
            """
            try:
                pyvb.DmmFunction(dmm_function)
            except Exception:
                try:
                    dmm_function = pyvb.DmmFunction[dmm_function.upper()]
                except Exception:
                    raise ValueError(
                        "DMM Function may be 0-5, 'DC_VOLTS'," +
                        " 'AC_VOLTS', 'DC_CURRENT', 'AC_CURRENT'," +
                        " 'RESISTANCE' or 'DIODE'")
            return dmm_function

        def validate_auto_range_terminal(self, auto_range_terminal):
            """ Check value for choosing the auto range terminal for
            DC current measurement

            :param auto_range_terminal: Terminal to perform
                                        auto ranging (``'LOW'``
                                        or ``'HIGH'``)
            :type auto_range_terminal: int or str
            :return: Auto range terminal to pass to the instrument
            :rtype: int
            """
            try:
                pyvb.DmmCurrentTerminal(auto_range_terminal)
            except Exception:
                try:
                    auto_range_terminal = pyvb.DmmCurrentTerminal[
                        auto_range_terminal.upper()]
                except Exception:
                    raise ValueError(
                        "Current Auto Range Terminal may be 0, 1," +
                        " 'LOW' or 'HIGH'")
            return auto_range_terminal

        def configure_measurement(self, dmm_function, auto_range=True,
                                  manual_range=1.0):
            """ Configure Instrument to take a DMM measurement

            :param dmm_function:DMM function index or name:

                - ``'DC_VOLTS'``, ``'AC_VOLTS'``
                - ``'DC_CURRENT'``, ``'AC_CURRENT'``
                - ``'RESISTANCE'``
                - ``'DIODE'``

            :type dmm_function: int or str
            :param bool auto_range: Enable/Disable auto ranging
            :param float manual_range: Manually set measurement range
            """
            dmm_function = self.validate_dmm_function(dmm_function)
            auto_range = strict_discrete_set(auto_range, [True, False])
            if auto_range is False:
                manual_range = self.validate_range(dmm_function, range)
            self.dmm.configure_measurement(
                dmm_function, auto_range=auto_range, manual_range=manual_range)

        def configure_dc_voltage(self, dmm_input_resistance):
            """ Configure DC voltage input resistance

            :param dmm_input_resistance: Input resistance (``'TEN_MEGA_OHM'``
                                         or ``'TEN_GIGA_OHM'``)
            :type dmm_input_resistance: int or str
            """
            try:
                pyvb.DmmInputResistance(dmm_input_resistance)
            except Exception:
                try:
                    dmm_input_resistance = pyvb.DmmInputResistance[
                        dmm_input_resistance.upper()]
                except Exception:
                    raise ValueError(
                        "Input Resistance may be 0, 1," +
                        " 'TEN_MEGA_OHM' or 'TEN_GIGA_OHM'")
            self.dmm.configure_dc_voltage(dmm_input_resistance)

        def configure_dc_current(self, auto_range_terminal):
            """ Configure auto rage terminal for DC current measurement

            :param auto_range_terminal: Terminal to perform auto ranging
                                        (``'LOW'`` or ``'HIGH'``)
            """
            auto_range_terminal = self.validate_auto_range_terminal(
                auto_range_terminal)
            self.dmm.configure_dc_current(auto_range_terminal)

        def configure_ac_current(self, auto_range_terminal):
            """ Configure auto rage terminal for AC current measurement

            :param auto_range_terminal: Terminal to perform auto ranging
                                        (``'LOW'`` or ``'HIGH'``)
            """
            auto_range_terminal = self.validate_auto_range_terminal(
                auto_range_terminal)
            self.dmm.configure_ac_current(auto_range_terminal)

        def query_measurement(self):
            """ Query DMM measurement settings from the instrument

            :return: Auto range, range data
            :rtype: (bool, float)
            """
            return self.dmm.query_measurement(0)

        def query_dc_voltage(self):
            """ Indicates input resistance setting for DC voltage measurement
            """
            self.dmm.query_dc_voltage()

        def query_dc_current(self):
            """ Indicates auto range terminal for DC current measurement
            """
            self.dmm.query_dc_current()

        def query_ac_current(self):
            """ Indicates auto range terminal for AC current measurement
            """
            self.dmm.query_ac_current()

        def read(self):
            """ Read measurement value from the instrument

            :return: Measurement value
            :rtype: float
            """
            self.dmm.read()

        def reset_instrument(self):
            """ Reset the DMM module to defaults
            """
            self.dmm.reset_instrument()

    class FunctionGenerator(VirtualBenchInstrument):
        """ Represents Function Generator (FGEN) Module of Virtual
        Bench device.
        """
        def __init__(self, virtualbench, reset, vb_name=''):
            """ Acquire FGEN module

            :param virtualbench: Instance of the VirtualBench class
            :type virtualbench: VirtualBench
            :param reset: Resets the instrument
            :type reset: bool
            """
            super().__init__(
                virtualbench.acquire_function_generator,
                reset, 'fgen', vb_name)
            self.fgen = self._instrument_handle

            self._waveform_functions = {"SINE": 0, "SQUARE": 1,
                                        "TRIANGLE/RAMP": 2, "DC": 3}
            # self._waveform_functions_index = {
            # v: k for k, v in self._waveform_functions.items()}
            self._max_frequency = {"SINE": 20000000, "SQUARE": 5000000,
                                   "TRIANGLE/RAMP": 1000000, "DC": 20000000}

        def configure_standard_waveform(self, waveform_function, amplitude,
                                        dc_offset, frequency, duty_cycle):
            """ Configures the instrument to output a standard waveform.
            Check instrument manual for maximum ratings which depend on load.

            :param waveform_function: Waveform function (``"SINE", "SQUARE",
                                      "TRIANGLE/RAMP", "DC"``)
            :type waveform_function: int or str
            :param amplitude: Amplitude in volts
            :type amplitude: float
            :param dc_offset: DC offset in volts
            :type dc_offset: float
            :param frequency: Frequency in Hz
            :type frequency: float
            :param duty_cycle: Duty cycle in %
            :type duty_cycle: int
            """
            waveform_function = strict_discrete_set(
                waveform_function.upper(), self._waveform_functions)
            max_frequency = self._max_frequency[waveform_function.upper()]
            waveform_function = self._waveform_functions[
                waveform_function.upper()]
            amplitude = strict_range(amplitude, (0, 24))
            dc_offset = strict_range(dc_offset, (-12, 12))
            if (amplitude/2 + abs(dc_offset)) > 12:
                raise ValueError(
                    "Amplitude and DC Offset may not exceed +/-12V")
            duty_cycle = strict_range(duty_cycle, (0, 100))
            frequency = strict_range(frequency, (0, max_frequency))
            self.fgen.configure_standard_waveform(
                waveform_function, amplitude, dc_offset, frequency, duty_cycle)

        def configure_arbitrary_waveform(self, waveform, sample_period):
            """ Configures the instrument to output a waveform. The waveform is
            output either after the end of the current waveform if output
            is enabled, or immediately after output is enabled.

            :param waveform: Waveform as list of values
            :type waveform: list
            :param sample_period: Time between two waveform points
                                  (maximum of 125MS/s, which equals 80ns)
            :type sample_period: float
            """
            strict_range(len(waveform), (1, 1e6))  # 1MS
            sample_period = strict_range(sample_period, (8e-8, 1))
            self.fgen.configure_arbitrary_waveform(waveform, sample_period)

        def configure_arbitrary_waveform_gain_and_offset(self, gain,
                                                         dc_offset):
            """ Configures the instrument to output an arbitrary waveform with
            a specified gain and offset value. The waveform is output either
            after the end of the current waveform if output is enabled, or
            immediately after output is enabled.

            :param gain: Gain, multiplier of waveform values
            :type gain: float
            :param dc_offset: DC offset in volts
            :type dc_offset: float
            """
            dc_offset = strict_range(dc_offset, (-12, 12))
            self.fgen.configure_arbitrary_waveform_gain_and_offset(
                gain, dc_offset)

        @property
        def filter(self):
            ''' Enables or disables the filter on the instrument.

            :param bool enable_filter: Enable/Disable filter
            '''
            return self.fgen.query_filter

        @filter.setter
        def filter(self, enable_filter):
            enable_filter = strict_discrete_set(enable_filter, [True, False])
            self.fgen.enable_filter(enable_filter)

        def query_waveform_mode(self):
            """ Indicates whether the waveform output by the instrument is a
            standard or arbitrary waveform.

            :return: Waveform mode
            :rtype: enum
            """
            return self.fgen.query_waveform_mode()

        def query_standard_waveform(self):
            """ Returns the settings for a standard waveform generation.

            :return: Waveform function, amplitude, dc_offset, frequency,
                     duty_cycle
            :rtype: (enum, float, float, float, int)
            """
            return self.fgen.query_standard_waveform()

        def query_arbitrary_waveform(self):
            """ Returns the samples per second for arbitrary waveform
            generation.

            :return: Samples per second
            :rtype: int
            """
            return self.fgen.query_arbitrary_waveform()

        def query_arbitrary_waveform_gain_and_offset(self):
            """ Returns the settings for arbitrary waveform generation that
            includes gain and offset settings.

            :return: Gain, DC offset
            :rtype: (float, float)
            """
            return self.fgen.query_arbitrary_waveform_gain_and_offset()

        def query_generation_status(self):
            """ Returns the status of waveform generation on the instrument.

            :return: Status
            :rtype: enum
            """
            return self.fgen.query_generation_status()

        def run(self):
            ''' Transitions the session from the Stopped state to the Running
            state.
            '''
            log.info("%s START" % self.name)
            self.fgen.run()

        def self_calibrate(self):
            '''Performs offset nulling calibration on the device. You must run
            FGEN Initialize prior to running this method.
            '''
            self.fgen.self_calibrate()

        def stop(self):
            ''' Transitions the acquisition from either the Triggered or
            Running state to the Stopped state.
            '''
            log.info("%s STOP" % self.name)
            self.fgen.stop()

        def reset_instrument(self):
            ''' Resets the session configuration to default values, and resets
            the device and driver software to a known state.
            '''
            self.fgen.reset_instrument()

    class MixedSignalOscilloscope(VirtualBenchInstrument):
        """ Represents Mixed Signal Oscilloscope (MSO) Module of Virtual Bench
        device. Allows to measure oscilloscope data from analog and digital
        channels.

        Methods from pyvirtualbench not implemented in pymeasure yet:

            - ``enable_digital_channels``
            - ``configure_digital_threshold``
            - ``configure_advanced_digital_timing``
            - ``configure_state_mode``
            - ``configure_digital_edge_trigger``
            - ``configure_digital_pattern_trigger``
            - ``configure_digital_glitch_trigger``
            - ``configure_digital_pulse_width_trigger``
            - ``query_digital_channel``
            - ``query_enabled_digital_channels``
            - ``query_digital_threshold``
            - ``query_advanced_digital_timing``
            - ``query_state_mode``
            - ``query_digital_edge_trigger``
            - ``query_digital_pattern_trigger``
            - ``query_digital_glitch_trigger``
            - ``query_digital_pulse_width_trigger``
            - ``read_digital_u64``
        """

        def __init__(self, virtualbench, reset, vb_name=''):
            """ Acquire MSO module

            :param virtualbench: Instance of the VirtualBench class
            :type virtualbench: VirtualBench
            :param reset: Resets the instrument
            :type reset: bool
            """
            super().__init__(
                virtualbench.acquire_mixed_signal_oscilloscope,
                reset, 'mso', vb_name)
            self.mso = self._instrument_handle

        @staticmethod
        def validate_trigger_instance(trigger_instance):
            """ Check if ``trigger_instance`` is a valid choice

            :param trigger_instance: Trigger instance (``'A'`` or ``'B'``)
            :type trigger_instance: int or str
            :return: Trigger instance
            :rtype: int
            """
            try:
                pyvb.MsoTriggerInstance(trigger_instance)
            except Exception:
                try:
                    trigger_instance = pyvb.MsoTriggerInstance[
                        trigger_instance.upper()]
                except Exception:
                    raise ValueError(
                        "Trigger Instance may be 0, 1, 'A' or 'B'")
            return trigger_instance

        def validate_channel(self, channel):
            """ Check if ``channel`` is a correct specification

            :param str channel: Channel string
            :return: Channel string
            :rtype: str
            """
            def error(channel=channel):
                raise ValueError(
                    "Channel specification {0} is not valid!".format(channel))
            channels = self._vb_handle.expand_channel_string(channel)[0]
            channels = channels.split(', ')
            return_value = []
            for channel in channels:
                # split off lines by last '/'
                try:
                    (device, channel) = re.match(
                        r'(.*)(?:/)(.+)', channel).groups()
                except Exception:
                    error()
                # validate numbers in range 1-2
                if not int(channel) in range(1, 3):
                    error()
                # validate device name: either 'mso' or 'device_name/mso'
                if device == 'mso':
                    pass
                else:
                    try:
                        device = re.match(
                            r'(VB[0-9]{4}-[0-9a-zA-Z]{7})(?:/)(.+)',
                            device).groups()[0]
                    except Exception:
                        error()
                    # device_name has to match
                    if not device == self._device_name:
                        error()
                # constructing line references for output
                return_value.append('mso/' + channel)

            return_value = ', '.join(return_value)
            return_value = self._vb_handle.collapse_channel_string(
                return_value)[0]  # drop number of channels
            return return_value

        # --------------------------
        # Configure Instrument
        # --------------------------

        def auto_setup(self):
            """ Automatically configure the instrument
            """
            self.mso.auto_setup()

        def configure_analog_channel(self, channel, enable_channel,
                                     vertical_range, vertical_offset,
                                     probe_attenuation, vertical_coupling):
            """ Configure analog measurement channel

            :param str channel: Channel string
            :param bool enable_channel: Enable/Disable channel
            :param float vertical_range: Vertical measurement range (0V - 20V),
                the instrument discretizes to these ranges:
                ``[20, 10, 5, 2, 1, 0.5, 0.2, 0.1, 0.05]``
                which are 5x the values shown in the native UI.
            :param float vertical_offset: Vertical offset to correct for
                                          (inverted compared to VB native UI,
                                          -20V - +20V, resolution 0.1mV)
            :param probe_attenuation: Probe attenuation (``'ATTENUATION_10X'``
                                      or ``'ATTENUATION_1X'``)
            :type probe_attenuation: int or str
            :param vertical_coupling: Vertical coupling (``'AC'`` or ``'DC'``)
            :type vertical_coupling: int or str
            """
            channel = self.validate_channel(channel)
            enable_channel = strict_discrete_set(
                enable_channel, [True, False])
            vertical_range = strict_range(vertical_range, (0, 20))
            vertical_offset = strict_discrete_range(
                vertical_offset, [-20, 20], 1e-4
            )
            try:
                pyvb.MsoProbeAttenuation(probe_attenuation)
            except Exception:
                try:
                    probe_attenuation = pyvb.MsoProbeAttenuation[
                        probe_attenuation.upper()]
                except Exception:
                    raise ValueError(
                        "Probe Attenuation may be 1, 10," +
                        " 'ATTENUATION_10X' or 'ATTENUATION_1X'")
            try:
                pyvb.MsoCoupling(vertical_coupling)
            except Exception:
                try:
                    vertical_coupling = pyvb.MsoCoupling[
                        vertical_coupling.upper()]
                except Exception:
                    raise ValueError(
                        "Probe Attenuation may be 0, 1, 'AC' or 'DC'")

            self.mso.configure_analog_channel(
                channel, enable_channel, vertical_range, vertical_offset,
                probe_attenuation, vertical_coupling)

        def configure_analog_channel_characteristics(self, channel,
                                                     input_impedance,
                                                     bandwidth_limit):
            """ Configure electrical characteristics of the specified channel

            :param str channel: Channel string
            :param input_impedance: Input Impedance (``'ONE_MEGA_OHM'`` or
                                    ``'FIFTY_OHMS'``)
            :type input_impedance: int or str
            :param int bandwidth_limit: Bandwidth limit (100MHz or 20MHz)
            """
            channel = self.validate_channel(channel)
            try:
                pyvb.MsoInputImpedance(input_impedance)
            except Exception:
                try:
                    input_impedance = pyvb.MsoInputImpedance[
                        input_impedance.upper()]
                except Exception:
                    raise ValueError(
                        "Probe Attenuation may be 0, 1," +
                        " 'ONE_MEGA_OHM' or 'FIFTY_OHMS'")
            bandwidth_limit = strict_discrete_set(
                bandwidth_limit, [100000000, 20000000])  # 100 Mhz or 20Mhz
            self.mso.configure_analog_channel_characteristics(
                channel, input_impedance, bandwidth_limit)

        def configure_timing(self, sample_rate, acquisition_time,
                             pretrigger_time, sampling_mode):
            """ Configure timing settings of the MSO

            :param int sample_rate: Sample rate (15.26kS - 1GS)
            :param float acquisition_time: Acquisition time (1ns - 68.711s)
            :param float pretrigger_time: Pretrigger time (0s - 10s)
            :param sampling_mode: Sampling mode (``'SAMPLE'`` or
                                  ``'PEAK_DETECT'``)
            """
            sample_rate = strict_range(sample_rate, (15260, 1e9))
            acquisition_time = strict_discrete_range(
                acquisition_time, (1e-09, 68.711), 1e-09)
            # acquisition is also limited by buffer size,
            # which depends on sample rate as well as acquisition time
            pretrigger_time = strict_range(pretrigger_time, (0, 10))
            try:
                pyvb.MsoSamplingMode(sampling_mode)
            except Exception:
                try:
                    sampling_mode = pyvb.MsoSamplingMode[sampling_mode.upper()]
                except Exception:
                    raise ValueError(
                        "Sampling Mode may be 0, 1, 'SAMPLE' or 'PEAK_DETECT'")

            self.mso.configure_timing(
                sample_rate, acquisition_time, pretrigger_time, sampling_mode)

        def configure_immediate_trigger(self):
            """ Configures a trigger to immediately activate on the specified
            channels after the pretrigger time has expired.
            """
            self.mso.configure_immediate_trigger()

        def configure_analog_edge_trigger(self, trigger_source, trigger_slope,
                                          trigger_level, trigger_hysteresis,
                                          trigger_instance):
            """ Configures a trigger to activate on the specified source when
            the analog edge reaches the specified levels.

            :param str trigger_source: Channel string
            :param trigger_slope: Trigger slope (``'RISING'``, ``'FALLING'``
                                  or ``'EITHER'``)
            :type trigger_slope: int or str
            :param float trigger_level: Trigger level
            :param float trigger_hysteresis: Trigger hysteresis
            :param trigger_instance: Trigger instance
            :type trigger_instance: int or str
            """
            trigger_source = self.validate_channel(trigger_source)
            try:
                pyvb.EdgeWithEither(trigger_slope)
            except Exception:
                try:
                    trigger_slope = pyvb.EdgeWithEither[trigger_slope.upper()]
                except Exception:
                    raise ValueError(
                        "Trigger Slope may be 0, 1, 2, 'RISING'," +
                        " 'FALLING' or 'EITHER'")
            trigger_instance = self.validate_trigger_instance(trigger_instance)
            self.mso.configure_analog_edge_trigger(
                trigger_source, trigger_slope, trigger_level,
                trigger_hysteresis, trigger_instance)

        def configure_analog_pulse_width_trigger(self, trigger_source,
                                                 trigger_polarity,
                                                 trigger_level,
                                                 comparison_mode, lower_limit,
                                                 upper_limit,
                                                 trigger_instance):
            """ Configures a trigger to activate on the specified source when
            the analog edge reaches the specified levels within a specified
            window of time.

            :param str trigger_source: Channel string
            :param trigger_polarity: Trigger slope (``'POSITIVE'`` or
                                     ``'NEGATIVE'``)
            :type trigger_polarity: int or str
            :param float trigger_level: Trigger level
            :param comparison_mode: Mode of compariosn (
                                    ``'GREATER_THAN_UPPER_LIMIT'``,
                                    ``'LESS_THAN_LOWER_LIMIT'``,
                                    ``'INSIDE_LIMITS'`` or
                                    ``'OUTSIDE_LIMITS'``)
            :type comparison_mode: int or str
            :param float lower_limit: Lower limit
            :param float upper_limit: Upper limit
            :param trigger_instance: Trigger instance
            :type trigger_instance: int or str
            """
            trigger_source = self.validate_channel(trigger_source)
            try:
                pyvb.MsoTriggerPolarity(trigger_polarity)
            except Exception:
                try:
                    trigger_polarity = pyvb.MsoTriggerPolarity[
                        trigger_polarity.upper()]
                except Exception:
                    raise ValueError(
                        "Comparison Mode may be 0, 1, 2, 3," +
                        " 'GREATER_THAN_UPPER_LIMIT'," +
                        " 'LESS_THAN_LOWER_LIMIT'," +
                        " 'INSIDE_LIMITS' or 'OUTSIDE_LIMITS'")
            try:
                pyvb.MsoComparisonMode(comparison_mode)
            except Exception:
                try:
                    comparison_mode = pyvb.MsoComparisonMode[
                        comparison_mode.upper()]
                except Exception:
                    raise ValueError(
                        "Trigger Polarity may be 0, 1," +
                        " 'POSITIVE' or 'NEGATIVE'")
            trigger_instance = self.validate_trigger_instance(trigger_instance)
            self.mso.configure_analog_pulse_width_trigger(
                trigger_source, trigger_polarity, trigger_level,
                comparison_mode, lower_limit, upper_limit, trigger_instance)

        def configure_trigger_delay(self, trigger_delay):
            """ Configures the amount of time to wait after a trigger condition
            is met before triggering.

                :param float trigger_delay: Trigger delay (0s - 17.1799s)
            """
            self.mso.configure_trigger_delay(trigger_delay)

        def query_analog_channel(self, channel):
            """ Indicates the vertical configuration of the specified channel.

            :return: Channel enabled, vertical range, vertical offset,
                     probe attenuation, vertical coupling
            :rtype: (bool, float, float, enum, enum)
            """
            channel = self.validate_channel(channel)
            return self.mso.query_analog_channel(channel)

        def query_enabled_analog_channels(self):
            """ Returns String of enabled analog channels.

            :return: Enabled analog channels
            :rtype: str
            """
            return self.mso.query_enabled_analog_channels()

        def query_analog_channel_characteristics(self, channel):
            """ Indicates the properties that control the electrical
            characteristics of the specified channel.
            This method returns an error if too much power is
            applied to the channel.

                :return: Input impedance, bandwidth limit
                :rtype: (enum, float)
            """
            return self.mso.query_analog_channel_characteristics(channel)

        def query_timing(self):
            """ Indicates the timing configuration of the MSO.
            Call directly before measurement to read the actual timing
            configuration and write it to the corresponding class variables.
            Necessary to interpret the measurement data, since it contains no
            time information.

            :return: Sample rate, acquisition time, pretrigger time,
                     sampling mode
            :rtype: (float, float, float, enum)
            """
            (self.sample_rate, self.acquisition_time,
             self.pretrigger_time,
             self.sampling_mode) = self.mso.query_timing()
            return (self.sample_rate, self.acquisition_time,
                    self.pretrigger_time, self.sampling_mode)

        def query_trigger_type(self, trigger_instance):
            """ Indicates the trigger type of the specified instance.

            :param trigger_instance: Trigger instance (``'A'`` or ``'B'``)
            :return: Trigger type
            :rtype: str
            """
            return self.mso.query_trigger_type()

        def query_analog_edge_trigger(self, trigger_instance):
            """ Indicates the analog edge trigger configuration of the
            specified instance.

            :return: Trigger source, trigger slope, trigger level, trigger
                     hysteresis
            :rtype: (str, enum, float, float)
            """
            trigger_instance = self.validate_trigger_instance(trigger_instance)
            return self.mso.query_analog_edge_trigger(trigger_instance)

        def query_trigger_delay(self):
            """ Indicates the trigger delay setting of the MSO.

            :return: Trigger delay
            :rtype: float
            """
            return self.mso.query_trigger_delay()

        def query_analog_pulse_width_trigger(self, trigger_instance):
            """ Indicates the analog pulse width trigger configuration of the
            specified instance.

            :return: Trigger source, trigger polarity, trigger level,
                     comparison mode, lower limit, upper limit
            :rtype: (str, enum, float, enum, float, float)
            """
            trigger_instance = self.validate_trigger_instance(trigger_instance)
            return self.mso.query_analog_pulse_width_trigger(trigger_instance)

        def query_acquisition_status(self):
            """ Returns the status of a completed or ongoing acquisition.
            """
            return self.mso.query_acquisition_status()

        # --------------------------
        # Measurement Control
        # --------------------------

        def run(self, autoTrigger=True):
            """ Transitions the acquisition from the Stopped state to the
            Running state. If the current state is Triggered, the
            acquisition is first transitioned to the Stopped state before
            transitioning to the Running state. This method returns an
            error if too much power is applied to any enabled channel.

            :param bool autoTrigger: Enable/Disable auto triggering
            """
            self.mso.run(autoTrigger)

        def force_trigger(self):
            """ Causes a software-timed trigger to occur after the pretrigger
            time has expired.
            """
            self.mso.force_trigger()

        def stop(self):
            """ Transitions the acquisition from either the Triggered or
            Running state to the Stopped state.
            """
            self.mso.stop()

        def read_analog_digital_u64(self):
            """ Transfers data from the instrument as long as the acquisition
            state is Acquisition Complete. If the state is either Running or
            Triggered, this method will wait until the state transitions to
            Acquisition Complete. If the state is Stopped, this method
            returns an error.

            :return: Analog data out, analog data stride, analog t0,
                     digital data out, digital timestamps out, digital t0,
                     trigger timestamp, trigger reason
            :rtype: (list, int, pyvb.Timestamp, list, list, pyvb.Timestamp,
                    pyvb.Timestamp, enum)
            """
            return self.mso.read_analog_digital_u64()

        def read_analog_digital_dataframe(self):
            """ Transfers data from the instrument and returns a pandas
            dataframe of the analog measurement data, including time
            coordinates

            :return: Dataframe with time and measurement data
            :rtype: pd.DataFrame
            """
            (analog_data_out, analog_data_stride
             # , analog_t0, digital_data_out, digital_timestamps_out,
             # digital_t0, trigger_timestamp, trigger_reason
             ) = self.read_analog_digital_u64()[0:2]

            number_of_samples = int(self.sample_rate *
                                    self.acquisition_time) + 1
            if not number_of_samples == (len(analog_data_out) /
                                         analog_data_stride):
                # try updating timing parameters
                self.query_timing()
                number_of_samples = int(self.sample_rate *
                                        self.acquisition_time) + 1
                if not number_of_samples == (len(analog_data_out) /
                                             analog_data_stride):
                    raise ValueError(
                        "Length of Analog Data does not match" +
                        " Timing Parameters")

            pretrigger_samples = int(self.sample_rate * self.pretrigger_time)
            times = (
                list(range(-pretrigger_samples, 0))
                + list(range(0, number_of_samples - pretrigger_samples)))
            times = [list(map(lambda x: x*1/self.sample_rate, times))]

            np_array = np.array(analog_data_out)
            np_array = np.split(np_array, analog_data_stride)
            np_array = np.append(np.array(times), np_array, axis=0)
            np_array = np.transpose(np_array)
            return pd.DataFrame(data=np_array)

        def reset_instrument(self):
            """ Resets the session configuration to default values, and resets
            the device and driver software to a known state.
            """
            self.mso.reset()

    class PowerSupply(VirtualBenchInstrument):
        """ Represents Power Supply (PS) Module of Virtual Bench device
        """
        def __init__(self, virtualbench, reset, vb_name=''):
            """ Acquire PS module

            :param virtualbench: Instance of the VirtualBench class
            :type virtualbench: VirtualBench
            :param reset: Resets the instrument
            :type reset: bool
            """
            super().__init__(
                virtualbench.acquire_power_supply,
                reset, 'ps', vb_name)
            self.ps = self._instrument_handle

        def validate_channel(self, channel, current=False, voltage=False):
            """ Check if channel string is valid and if output current/voltage
            are within the output ranges of the channel

            :param channel: Channel string (``"ps/+6V","ps/+25V","ps/-25V"``)
            :type channel: str
            :param current: Current output, defaults to False
            :type current: bool, optional
            :param voltage: Voltage output, defaults to False
            :type voltage: bool, optional
            :return: channel or channel, current & voltage
            :rtype: str or (str, float, float)
            """
            if current is False and voltage is False:
                return strict_discrete_set(
                    channel, ["ps/+6V", "ps/+25V", "ps/-25V"])
            else:
                channel = strict_discrete_set(
                    channel, ["ps/+6V", "ps/+25V", "ps/-25V"])
                if channel == "ps/+6V":
                    current_range = (0, 1)
                    voltage_range = (0, 6)
                else:
                    current_range = (0, 5)
                    voltage_range = (0, 25)
                    if channel == "ps/-25V":
                        voltage_range = (0, -25)
                current = strict_discrete_range(current, current_range, 1e-3)
                voltage = strict_discrete_range(voltage, voltage_range, 1e-3)
                return (channel, current, voltage)

        def configure_voltage_output(self, channel, voltage_level,
                                     current_limit):
            ''' Configures a voltage output on the specified channel. This
            method should be called once for every channel you want to
            configure to output voltage.
            '''
            (channel, current_limit, voltage_level) = self.validate_channel(
                channel, current_limit, voltage_level)
            self.ps.configure_voltage_output(
                channel, voltage_level, current_limit)

        def configure_current_output(self, channel, current_level,
                                     voltage_limit):
            ''' Configures a current output on the specified channel. This
            method should be called once for every channel you want to
            configure to output current.
            '''
            (channel, current_level, voltage_limit) = self.validate_channel(
                channel, current_level, voltage_limit)
            self.ps.configure_current_output(
                channel, current_level, voltage_limit)

        def query_voltage_output(self, channel):
            ''' Indicates the voltage output settings on the specified channel.
            '''
            channel = self.validate_channel(channel)
            return self.ps.query_voltage_output(channel)

        def query_current_output(self, channel):
            ''' Indicates the current output settings on the specified channel.
            '''
            channel = self.validate_channel(channel)
            return self.ps.query_current_output(channel)

        @property
        def outputs_enabled(self):
            ''' Enables or disables all outputs on all channels of the
            instrument.

            :param bool enable_outputs: Enable/Disable outputs
            '''
            return self.ps.query_outputs_enabled()

        @outputs_enabled.setter
        def outputs_enabled(self, enable_outputs):
            enable_outputs = strict_discrete_set(
                enable_outputs, [True, False])
            log.info("%s Output %s." % (self.name, enable_outputs))
            self.ps.enable_all_outputs(enable_outputs)

        @property
        def tracking(self):
            ''' Enables or disables tracking between the positive and negative
            25V channels. If enabled, any configuration change on the
            positive 25V channel is mirrored to the negative 25V channel,
            and any writes to the negative 25V channel are ignored.

            :param bool enable_tracking: Enable/Disable tracking
            '''
            return self.ps.query_tracking()

        @tracking.setter
        def tracking(self, enable_tracking):
            enable_tracking = strict_discrete_set(
                enable_tracking, [True, False])
            self.ps.enable_tracking(enable_tracking)

        def read_output(self, channel):
            ''' Reads the voltage and current levels and outout mode of the
            specified channel.
            '''
            channel = self.validate_channel(channel)
            return self.ps.read_output()

        def reset_instrument(self):
            ''' Resets the session configuration to default values, and resets
            the device and driver software to a known state.
            '''
            self.ps.reset_instrument()



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import sys
import logging
import time
import traceback
from logging.handlers import QueueHandler
from importlib.machinery import SourceFileLoader
from queue import Queue

from .listeners import Recorder
from .procedure import Procedure, ProcedureWrapper
from .results import Results
from ..log import TopicQueueHandler
from ..thread import StoppableThread

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

try:
    import zmq
    import cloudpickle
except ImportError:
    zmq = None
    cloudpickle = None
    log.warning("ZMQ and cloudpickle are required for TCP communication")


class Worker(StoppableThread):
    """ Worker runs the procedure and emits information about
    the procedure and its status over a ZMQ TCP port. In a child
    thread, a Recorder is run to write the results to
    """

    def __init__(self, results, log_queue=None, log_level=logging.INFO, port=None):
        """ Constructs a Worker to perform the Procedure
        defined in the file at the filepath
        """
        super().__init__()

        self.port = port
        if not isinstance(results, Results):
            raise ValueError("Invalid Results object during Worker construction")
        self.results = results
        self.results.procedure.check_parameters()
        self.results.procedure.status = Procedure.QUEUED

        self.recorder = None
        self.recorder_queue = Queue()

        self.monitor_queue = Queue()
        if log_queue is None:
            log_queue = Queue()
        self.log_queue = log_queue
        self.log_level = log_level

        self.context = None
        self.publisher = None

    def join(self, timeout=0):
        try:
            super().join(timeout)
        except (KeyboardInterrupt, SystemExit):
            log.warning("User stopped Worker join prematurely")
            self.stop()
            super().join(0)

    def emit(self, topic, record):
        """ Emits data of some topic over TCP """
        log.debug("Emitting message: %s %s", topic, record)

        try:
            self.publisher.send_serialized((topic, record), serialize=cloudpickle.dumps)
        except (NameError, AttributeError):
            pass  # No dumps defined
        if topic == 'results':
            self.recorder.handle(record)
        elif topic == 'status' or topic == 'progress':
            self.monitor_queue.put((topic, record))

    def handle_abort(self):
        log.exception("User stopped Worker execution prematurely")
        self.update_status(Procedure.ABORTED)

    def handle_error(self):
        log.exception("Worker caught an error on %r", self.procedure)
        traceback_str = traceback.format_exc()
        self.emit('error', traceback_str)
        self.update_status(Procedure.FAILED)

    def update_status(self, status):
        self.procedure.status = status
        self.emit('status', status)

    def shutdown(self):
        self.procedure.shutdown()

        if self.should_stop() and self.procedure.status == Procedure.RUNNING:
            self.update_status(Procedure.ABORTED)
        elif self.procedure.status == Procedure.RUNNING:
            self.update_status(Procedure.FINISHED)
            self.emit('progress', 100.)

        self.recorder.stop()
        self.monitor_queue.put(None)

    def run(self):
        global log
        log = logging.getLogger()
        log.setLevel(self.log_level)
        # log.handlers = []  # Remove all other handlers
        # log.addHandler(TopicQueueHandler(self.monitor_queue))
        # log.addHandler(QueueHandler(self.log_queue))
        log.info("Worker thread started")

        self.procedure = self.results.procedure

        self.recorder = Recorder(self.results, self.recorder_queue)
        self.recorder.start()

        #locals()[self.procedures_file] = __import__(self.procedures_file)

        # route Procedure methods & log
        self.procedure.should_stop = self.should_stop
        self.procedure.emit = self.emit

        if self.port is not None and zmq is not None:
            try:
                self.context = zmq.Context()
                log.debug("Worker ZMQ Context: %r" % self.context)
                self.publisher = self.context.socket(zmq.PUB)
                self.publisher.bind('tcp://*:%d' % self.port)
                log.info("Worker connected to tcp://*:%d" % self.port)
                time.sleep(0.01)
            except Exception:
                log.exception("couldn't connect to ZMQ context")

        log.info("Worker started running an instance of %r", self.procedure.__class__.__name__)
        self.update_status(Procedure.RUNNING)
        self.emit('progress', 0.)

        try:
            self.procedure.startup()
            self.procedure.execute()
        except (KeyboardInterrupt, SystemExit):
            self.handle_abort()
        except Exception:
            self.handle_error()
        finally:
            self.shutdown()
            self.stop()

    def __repr__(self):
        return "<%s(port=%s,procedure=%s,should_stop=%s)>" % (
            self.__class__.__name__, self.port,
            self.procedure.__class__.__name__,
            self.should_stop()
        )



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
from pymeasure.instruments import Instrument
from pymeasure.instruments.validators import strict_range
from .agilent33500 import Agilent33500


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Agilent33521A(Agilent33500):
    """Represents the Agilent 33521A Function/Arbitrary Waveform Generator.
    This documentation page shows only methods different from the parent class :doc:`Agilent33500 <agilent33500>`.

    """

    def __init__(self, adapter, **kwargs):
        super(Agilent33521A, self).__init__(
            adapter,
            **kwargs
        )

    frequency = Instrument.control(
        "FREQ?", "FREQ %f",
        """ A floating point property that controls the frequency of the output
        waveform in Hz, from 1 uHz to 30 MHz, depending on the specified function.
        Can be set. """,
        validator=strict_range,
        values=[1e-6, 30e+6],
    )

    arb_srate = Instrument.control(
        "FUNC:ARB:SRAT?", "FUNC:ARB:SRAT %f",
        """ An floating point property that sets the sample rate of the currently selected 
        arbitrary signal. Valid values are 1 µSa/s to 250 MSa/s. This can be set. """,
        validator=strict_range,
        values=[1e-6, 250e6],
    )


#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging

import os
import re
import sys
from copy import deepcopy
from importlib.machinery import SourceFileLoader
from datetime import datetime

import pandas as pd

from .procedure import Procedure, UnknownProcedure
from .parameters import Parameter

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def unique_filename(directory, prefix='DATA', suffix='', ext='csv',
                    dated_folder=False, index=True, datetimeformat="%Y-%m-%d"):
    """ Returns a unique filename based on the directory and prefix
    """
    now = datetime.now()
    directory = os.path.abspath(directory)
    if dated_folder:
        directory = os.path.join(directory, now.strftime('%Y-%m-%d'))
    if not os.path.exists(directory):
        os.makedirs(directory)
    if index:
        i = 1
        basename = "%s%s" % (prefix, now.strftime(datetimeformat))
        basepath = os.path.join(directory, basename)
        filename = "%s_%d%s.%s" % (basepath, i, suffix, ext)
        while os.path.exists(filename):
            i += 1
            filename = "%s_%d%s.%s" % (basepath, i, suffix, ext)
    else:
        basename = "%s%s%s.%s" % (prefix, now.strftime(datetimeformat), suffix, ext)
        filename = os.path.join(directory, basename)
    return filename


class CSVFormatter(logging.Formatter):
    """ Formatter of data results """

    def __init__(self, columns, delimiter=','):
        """Creates a csv formatter for a given list of columns (=header).

        :param columns: list of column names.
        :type columns: list
        :param delimiter: delimiter between columns.
        :type delimiter: str
        """
        super().__init__()
        self.columns = columns
        self.delimiter = delimiter

    def format(self, record):
        """Formats a record as csv.

        :param record: record to format.
        :type record: dict
        :return: a string
        """
        return self.delimiter.join('{}'.format(record[x]) for x in self.columns)

    def format_header(self):
        return self.delimiter.join(self.columns)


class Results(object):
    """ The Results class provides a convenient interface to reading and
    writing data in connection with a :class:`.Procedure` object.

    :cvar COMMENT: The character used to identify a comment (default: #)
    :cvar DELIMITER: The character used to delimit the data (default: ,)
    :cvar LINE_BREAK: The character used for line breaks (default \\n)
    :cvar CHUNK_SIZE: The length of the data chuck that is read

    :param procedure: Procedure object
    :param data_filename: The data filename where the data is or should be
                          stored
    """

    COMMENT = '#'
    DELIMITER = ','
    LINE_BREAK = "\n"
    CHUNK_SIZE = 1000

    def __init__(self, procedure, data_filename):
        if not isinstance(procedure, Procedure):
            raise ValueError("Results require a Procedure object")
        self.procedure = procedure
        self.procedure_class = procedure.__class__
        self.parameters = procedure.parameter_objects()
        self._header_count = -1

        self.formatter = CSVFormatter(columns=self.procedure.DATA_COLUMNS)

        if isinstance(data_filename, (list, tuple)):
            data_filenames, data_filename = data_filename, data_filename[0]
        else:
            data_filenames = [data_filename]

        self.data_filename = data_filename
        self.data_filenames = data_filenames

        if os.path.exists(data_filename):  # Assume header is already written
            self.reload()
            self.procedure.status = Procedure.FINISHED
            # TODO: Correctly store and retrieve status
        else:
            for filename in self.data_filenames:
                with open(filename, 'w') as f:
                    f.write(self.header())
                    f.write(self.labels())
            self._data = None

    def __getstate__(self):
        # Get all information needed to reconstruct procedure
        self._parameters = self.procedure.parameter_values()
        self._class = self.procedure.__class__.__name__
        module = sys.modules[self.procedure.__module__]
        self._package = module.__package__
        self._module = module.__name__
        self._file = module.__file__

        state = self.__dict__.copy()
        del state['procedure']
        del state['procedure_class']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        # Restore the procedure
        module = SourceFileLoader(self._module, self._file).load_module()
        cls = getattr(module, self._class)

        self.procedure = cls()
        self.procedure.set_parameters(self._parameters)
        self.procedure.refresh_parameters()

        self.procedure_class = cls

        del self._parameters
        del self._class
        del self._package
        del self._module
        del self._file

    def header(self):
        """ Returns a text header to accompany a datafile so that the procedure
        can be reconstructed
        """
        h = []
        procedure = re.search("'(?P<name>[^']+)'",
                              repr(self.procedure_class)).group("name")
        h.append("Procedure: <%s>" % procedure)
        h.append("Parameters:")
        for name, parameter in self.parameters.items():
            h.append("\t%s: %s" % (parameter.name, str(parameter).encode("unicode_escape").decode("utf-8")))
        h.append("Data:")
        self._header_count = len(h)
        h = [Results.COMMENT + l for l in h]  # Comment each line
        return Results.LINE_BREAK.join(h) + Results.LINE_BREAK

    def labels(self):
        """ Returns the columns labels as a string to be written
        to the file
        """
        return self.formatter.format_header() + Results.LINE_BREAK

    def format(self, data):
        """ Returns a formatted string containing the data to be written
        to a file
        """
        return self.formatter.format(data)

    def parse(self, line):
        """ Returns a dictionary containing the data from the line """
        data = {}
        items = line.split(Results.DELIMITER)
        for i, key in enumerate(self.procedure.DATA_COLUMNS):
            data[key] = items[i]
        return data

    @staticmethod
    def parse_header(header, procedure_class=None):
        """ Returns a Procedure object with the parameters as defined in the
        header text.
        """
        if procedure_class is not None:
            procedure = procedure_class()
        else:
            procedure = None

        header = header.split(Results.LINE_BREAK)
        procedure_module = None
        parameters = {}
        for line in header:
            if line.startswith(Results.COMMENT):
                line = line[1:]  # Uncomment
            else:
                raise ValueError("Parsing a header which contains "
                                 "uncommented sections")
            if line.startswith("Procedure"):
                regex = r"<(?:(?P<module>[^>]+)\.)?(?P<class>[^.>]+)>"
                search = re.search(regex, line)
                procedure_module = search.group("module")
                procedure_class = search.group("class")
            elif line.startswith("\t"):
                separator = ": "
                partitioned_line = line[1:].partition(separator)
                if partitioned_line[1] != separator:
                    raise Exception("Error partitioning header line %s." % line)
                else:
                    parameters[partitioned_line[0]] = partitioned_line[2]

        if procedure is None:
            if procedure_class is None:
                raise ValueError("Header does not contain the Procedure class")
            try:
                from importlib import import_module
                procedure_module = import_module(procedure_module)
                procedure_class = getattr(procedure_module, procedure_class)
                procedure = procedure_class()
            except ImportError:
                procedure = UnknownProcedure(parameters)
                log.warning("Unknown Procedure being used")
            except Exception as e:
                raise e

        # Fill the procedure with the parameters found
        for name, parameter in procedure.parameter_objects().items():
            if parameter.name in parameters:
                value = parameters[parameter.name]
                setattr(procedure, name, value)
            else:
                raise Exception("Missing '%s' parameter when loading '%s' class" % (
                    parameter.name, procedure_class))

        procedure.refresh_parameters()  # Enforce update of meta data
        return procedure

    @staticmethod
    def load(data_filename, procedure_class=None):
        """ Returns a Results object with the associated Procedure object and
        data
        """
        header = ""
        header_read = False
        header_count = 0
        with open(data_filename, 'r') as f:
            while not header_read:
                line = f.readline()
                if line.startswith(Results.COMMENT):
                    header += line.strip() + Results.LINE_BREAK
                    header_count += 1
                else:
                    header_read = True
        procedure = Results.parse_header(header[:-1], procedure_class)
        results = Results(procedure, data_filename)
        results._header_count = header_count
        return results

    @property
    def data(self):
        # Need to update header count for correct referencing
        if self._header_count == -1:
            self._header_count = len(
                self.header()[-1].split(Results.LINE_BREAK))
        if self._data is None or len(self._data) == 0:
            # Data has not been read
            try:
                self.reload()
            except Exception:
                # Empty dataframe
                self._data = pd.DataFrame(columns=self.procedure.DATA_COLUMNS)
        else:  # Concatenate additional data, if any, to already loaded data
            skiprows = len(self._data) + self._header_count
            chunks = pd.read_csv(
                self.data_filename,
                comment=Results.COMMENT,
                header=0,
                names=self._data.columns,
                chunksize=Results.CHUNK_SIZE, skiprows=skiprows, iterator=True
            )
            try:
                tmp_frame = pd.concat(chunks, ignore_index=True)
                # only append new data if there is any
                # if no new data, tmp_frame dtype is object, which override's
                # self._data's original dtype - this can cause problems plotting
                # (e.g. if trying to plot int data on a log axis)
                if len(tmp_frame) > 0:
                    self._data = pd.concat([self._data, tmp_frame],
                                           ignore_index=True)
            except Exception:
                pass  # All data is up to date
        return self._data

    def reload(self):
        """ Preforms a full reloading of the file data, neglecting
        any changes in the comments
        """
        chunks = pd.read_csv(
            self.data_filename,
            comment=Results.COMMENT,
            chunksize=Results.CHUNK_SIZE,
            iterator=True
        )
        try:
            self._data = pd.concat(chunks, ignore_index=True)
        except Exception:
            self._data = chunks.read()

    def __repr__(self):
        return "<{}(filename='{}',procedure={},shape={})>".format(
            self.__class__.__name__, self.data_filename,
            self.procedure.__class__.__name__,
            self.data.shape
        )



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
import logging.handlers
from logging.handlers import QueueHandler

from queue import Queue

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class QueueListener(logging.handlers.QueueListener):
    def is_alive(self):
        try:
            return self._thread.is_alive()
        except AttributeError:
            return False


def console_log(logger, level=logging.INFO, queue=None):
    """Create a console log handler. Return a scribe thread object."""
    if queue is None:
        queue = Queue()
    logger.setLevel(level)
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter(
        fmt='%(asctime)s: %(message)s (%(name)s, %(levelname)s)',
        datefmt='%I:%M:%S %p')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    scribe = Scribe(queue)
    return scribe


def file_log(logger, log_filename, level=logging.INFO, queue=None, **kwargs):
    """Create a file log handler. Return a scribe thread object."""
    if queue is None:
        queue = Queue()
    logger.setLevel(level)
    ch = logging.FileHandler(log_filename, **kwargs)
    ch.setLevel(level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    scribe = Scribe(queue)
    return scribe


class Scribe(QueueListener):
    """ Scribe class which logs records as retrieved from a queue to support consistent
    multi-process logging.

    :param queue: The multiprocessing queue which the scriber will listen to.
    """

    def __init__(self, queue):
        super().__init__(queue)

    def handle(self, record):
        logging.getLogger(record.name).handle(record)


def setup_logging(logger=None, console=False, console_level='INFO', filename=None,
                  file_level='DEBUG', queue=None, file_kwargs=None):
    """Setup logging for console and/or file logging. Returns a scribe thread object.
    Defaults to no logging."""
    if queue is None:
        queue = Queue()
    if logger is None:
        logger = logging.getLogger()
    if file_kwargs is None:
        file_kwargs = {}

    logger.handlers = []
    if console:
        console_log(logger, level=getattr(logging, console_level))
        logger.info('Set up console logging')
    if filename is not None:
        file_log(logger, filename, level=getattr(logging, file_level), **file_kwargs)
        logger.info('Set up file logging')

    scribe = Scribe(queue)
    return scribe


class TopicQueueHandler(QueueHandler):
    def __init__(self, queue, topic='log'):
        super().__init__(queue)
        self.topic = topic

    def prepare(self, record):
        return self.topic, record



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from pymeasure.instruments import Instrument, discreteTruncate, RangeException


class AnritsuMG3692C(Instrument):
    """ Represents the Anritsu MG3692C Signal Generator
    """
    power = Instrument.control(
        ":POWER?;", ":POWER %g dBm;",
        """ A floating point property that represents the output power
        in dBm. This property can be set. """
    )
    frequency = Instrument.control(
        ":FREQUENCY?;", ":FREQUENCY %e Hz;",
        """ A floating point property that represents the output frequency
        in Hz. This property can be set. """
    )

    def __init__(self, resourceName, **kwargs):
        super(AnritsuMG3692C, self).__init__(
            resourceName,
            "Anritsu MG3692C Signal Generator",
            **kwargs
        )

    @property
    def output(self):
        """ A boolean property that represents the signal output state.
        This property can be set to control the output.
        """
        return int(self.ask(":OUTPUT?")) == 1

    @output.setter
    def output(self, value):
        if value:
            self.write(":OUTPUT ON;")
        else:
            self.write(":OUTPUT OFF;")

    def enable(self):
        """ Enables the signal output.
        """
        self.output = True

    def disable(self):
        """ Disables the signal output.
        """
        self.output = False

    def shutdown(self):
        """ Shuts down the instrument, putting it in a safe state.
        """
        # TODO: Implement modulation
        self.modulation = False
        self.disable()



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
import time

import serial

from .serial import SerialAdapter


class PrologixAdapter(SerialAdapter):
    """ Encapsulates the additional commands necessary
    to communicate over a Prologix GPIB-USB Adapter,
    using the SerialAdapter.

    Each PrologixAdapter is constructed based on a serial port or
    connection and the GPIB address to be communicated to.
    Serial connection sharing is achieved by using the :meth:`.gpib`
    method to spawn new PrologixAdapters for different GPIB addresses.

    :param port: The Serial port name or a serial.Serial object
    :param address: Integer GPIB address of the desired instrument
    :param rw_delay: An optional delay to set between a write and read call for slow to respond instruments.
    :param preprocess_reply: optional callable used to preprocess strings
        received from the instrument. The callable returns the processed string.
    :param kwargs: Key-word arguments if constructing a new serial object

    :ivar address: Integer GPIB address of the desired instrument

    To allow user access to the Prologix adapter in Linux, create the file:
    :code:`/etc/udev/rules.d/51-prologix.rules`, with contents:

    .. code-block:: bash

        SUBSYSTEMS=="usb",ATTRS{idVendor}=="0403",ATTRS{idProduct}=="6001",MODE="0666"

    Then reload the udev rules with:

    .. code-block:: bash

        sudo udevadm control --reload-rules
        sudo udevadm trigger

    """

    def __init__(self, port, address=None, rw_delay=None, serial_timeout=0.5,
                 preprocess_reply=None, **kwargs):
        super().__init__(port, timeout=serial_timeout,
                         preprocess_reply=preprocess_reply, **kwargs)
        self.address = address
        self.rw_delay = rw_delay
        if not isinstance(port, serial.Serial):
            self.set_defaults()

    def set_defaults(self):
        """ Sets up the default behavior of the Prologix-GPIB
        adapter
        """
        self.write("++auto 0")  # Turn off auto read-after-write
        self.write("++eoi 1")  # Append end-of-line to commands
        self.write("++eos 2")  # Append line-feed to commands

    def ask(self, command):
        """ Ask the Prologix controller, include a forced delay for some instruments.

        :param command: SCPI command string to be sent to instrument
        """

        self.write(command)
        if self.rw_delay is not None:
            time.sleep(self.rw_delay)
        return self.read()

    def write(self, command):
        """ Writes the command to the GPIB address stored in the
        :attr:`.address`

        :param command: SCPI command string to be sent to the instrument
        """
        if self.address is not None:
            address_command = "++addr %d\n" % self.address
            self.connection.write(address_command.encode())
        command += "\n"
        self.connection.write(command.encode())

    def read(self):
        """ Reads the response of the instrument until timeout

        :returns: String ASCII response of the instrument
        """
        self.write("++read eoi")
        return b"\n".join(self.connection.readlines()).decode()

    def gpib(self, address, rw_delay=None):
        """ Returns and PrologixAdapter object that references the GPIB
        address specified, while sharing the Serial connection with other
        calls of this function

        :param address: Integer GPIB address of the desired instrument
        :param rw_delay: Set a custom Read/Write delay for the instrument
        :returns: PrologixAdapter for specific GPIB address
        """
        rw_delay = rw_delay or self.rw_delay
        return PrologixAdapter(self.connection, address, rw_delay=rw_delay)

    def wait_for_srq(self, timeout=25, delay=0.1):
        """ Blocks until a SRQ, and leaves the bit high

        :param timeout: Timeout duration in seconds
        :param delay: Time delay between checking SRQ in seconds
        """
        while int(self.ask("++srq")) != 1:  # TODO: Include timeout!
            time.sleep(delay)

    def __repr__(self):
        if self.address is not None:
            return "<PrologixAdapter(port='%s',address=%d)>" % (
                self.connection.port, self.address)
        else:
            return "<PrologixAdapter(port='%s')>" % self.connection.port



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#


from pymeasure.instruments import Instrument
from pymeasure.instruments.validators import truncated_range, strict_discrete_set, strict_range
from pyvisa.errors import VisaIOError


class AgilentE4980(Instrument):
    """Represents LCR meter E4980A/AL"""

    ac_voltage = Instrument.control(":VOLT:LEV?", ":VOLT:LEV %g",
                                    "AC voltage level, in Volts",
                                    validator=strict_range,
                                    values=[0, 20])
    
    ac_current = Instrument.control(":CURR:LEV?", ":CURR:LEV %g",
                                    "AC current level, in Amps",
                                    validator=strict_range,
                                    values=[0, 0.1])
    

    frequency = Instrument.control(":FREQ:CW?", ":FREQ:CW %g",
                                   "AC frequency (range depending on model), in Hertz",
                                   validator=strict_range,
                                   values=[20, 2e6])

    # FETCH? returns [A,B,state]: impedance returns only A,B
    impedance = Instrument.measurement(":FETCH?", 
                                       "Measured data A and B, according to :attr:`~.AgilentE4980.mode`",
                                       get_process=lambda x: x[:2])

    mode = Instrument.control("FUNCtion:IMPedance:TYPE?", "FUNCtion:IMPedance:TYPE %s",
                              """
Select quantities to be measured:
    
    * CPD: Parallel capacitance [F] and dissipation factor [number]
    * CPQ: Parallel capacitance [F] and quality factor [number]
    * CPG: Parallel capacitance [F] and parallel conductance [S]
    * CPRP: Parallel capacitance [F] and parallel resistance [Ohm]

    * CSD: Series capacitance [F] and dissipation factor [number]
    * CSQ: Series capacitance [F] and quality factor [number]
    * CSRS: Series capacitance [F] and series resistance [Ohm]

   * LPD: Parallel inductance [H] and dissipation factor [number]
   * LPQ: Parallel inductance [H] and quality factor [number]
   * LPG: Parallel inductance [H] and parallel conductance [S]
   * LPRP: Parallel inductance [H] and parallel resistance [Ohm]
   
    * LSD: Series inductance [H] and dissipation factor [number]
    * LSQ: Seriesinductance [H] and quality factor [number]
    * LSRS: Series inductance [H] and series resistance [Ohm]
    
    * RX: Resitance [Ohm] and reactance [Ohm]
    * ZTD: Impedance, magnitude [Ohm] and phase [deg]
    * ZTR: Impedance, magnitude [Ohm] and phase [rad]
    * GB: Conductance [S] and susceptance [S]
    * YTD: Admittance, magnitude [Ohm] and phase [deg]
    * YTR: Admittance magnitude [Ohm] and phase [rad]
""",
                              validator=strict_discrete_set,
                              values=["CPD", "CPQ", "CPG", "CPRP",
                                      "CSD", "CSQ", "CSRS",
                                      "LPD", "LPQ", "LPG", "LPRP",
                                      "LSD", "LSQ", "LSRS", 
                                      "RX", "ZTD", "ZTR", "GB", "YTD", "YTR",])


    trigger_source = Instrument.control("TRIG:SOUR?", "TRIG:SOUR %s",
                                            """
Select trigger source; accept the values:
    * HOLD: manual
    * INT: internal
    * BUS: external bus (GPIB/LAN/USB)
    * EXT: external connector""",
                                        validator=strict_discrete_set,
                                        values = ["HOLD", "INT", "BUS", "EXT"])

    def __init__(self, adapter, **kwargs):
        super(AgilentE4980, self).__init__(
            adapter, "Agilent E4980A/AL LCR meter", **kwargs
        )
        self.timeout = 30000
        # format: output ascii
        self.write("FORM ASC")

    def freq_sweep(self, freq_list, return_freq=False):
        """
        Run frequency list sweep using sequential trigger.

        :param freq_list: list of frequencies
        :param return_freq: if True, returns the frequencies read from the instrument
        
        Returns values as configured with :attr:`~.AgilentE4980.mode`
            """
        # manual, page 299
        # self.write("*RST;*CLS")
        self.write("TRIG:SOUR BUS")
        self.write("DISP:PAGE LIST")
        self.write("FORM ASC")
        #trigger in sequential mode
        self.write("LIST:MODE SEQ")
        lista_str = ",".join(['%e' % f for f in freq_list])
        self.write("LIST:FREQ %s" % lista_str)
        # trigger
        self.write("INIT:CONT ON")
        self.write(":TRIG:IMM")
        #wait for completed measurement
        #using the Error signal (there should be a better way)
        while 1:
            try:
                measured = self.values(":FETCh:IMPedance:FORMatted?")
                break
            except VisaIOError:
                pass
        #at the end return to manual trigger
        self.write(":TRIG:SOUR HOLD")
        # gets 4-ples of numbers, first two are data A and B
        a_data = [measured[_] for _ in range(0,  4*len(freq_list), 4)]
        b_data = [measured[_] for _ in range(1,  4*len(freq_list), 4)]
        if return_freq:
            read_freqs = self.values("LIST:FREQ?")
            return a_data, b_data, read_freqs
        else:
            return a_data, b_data

    # TODO: maybe refactor as property?
    def aperture(self, time=None, averages=1):
        """
        Set and get aperture.

        :param time: integration time as string: SHORT, MED, LONG (case insensitive); if None, get values
        :param averages: number of averages, numeric
        """
        if time is None:
            read_values = self.ask(":APER?").split(',')
            return read_values[0], int(read_values[1])
        else:
            if time.upper() in ["SHORT", "MED", "LONG"]:
                self.write(":APER {0}, {1}".format(time, averages))
            else:
                raise Exception("Time must be a string: SHORT, MED, LONG")



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

from pymeasure.instruments import Instrument
from pymeasure.instruments.validators import strict_discrete_set,\
    strict_range, joined_validators
from time import time
from pyvisa.errors import VisaIOError


# Capitalize string arguments to allow for better conformity with other WFG's
def capitalize_string(string: str, *args, **kwargs):
    return string.upper()


# Combine the capitalize function and validator
string_validator = joined_validators(capitalize_string, strict_discrete_set)


class Agilent33220A(Instrument):
    """Represents the Agilent 33220A Arbitrary Waveform Generator.

    .. code-block:: python

        # Default channel for the Agilent 33220A
        wfg = Agilent33220A("GPIB::10")

        wfg.shape = "SINUSOID"          # Sets a sine waveform
        wfg.frequency = 4.7e3           # Sets the frequency to 4.7 kHz
        wfg.amplitude = 1               # Set amplitude of 1 V
        wfg.offset = 0                  # Set the amplitude to 0 V

        wfg.burst = True                # Enable burst mode
        wfg.burst_ncycles = 10e3        # A burst will consist of 10 cycles
        wfg.burst_mode = "TRIGGERED"    # A burst will be applied on a trigger
        wfg.trigger_source = "BUS"      # A burst will be triggered on TRG*

        wfg.output = True               # Enable output of waveform generator
        wfg.trigger()                   # Trigger a burst
        wfg.wait_for_trigger()          # Wait until the triggering is finished
        wfg.beep()                      # "beep"

        wfg.check_errors()              # Get the error queue

    """

    def __init__(self, adapter, **kwargs):
        super(Agilent33220A, self).__init__(
            adapter,
            "Agilent 33220A Arbitrary Waveform generator",
            **kwargs
        )

    shape = Instrument.control(
        "FUNC?", "FUNC %s",
        """ A string property that controls the output waveform. Can be set to:
        SIN<USOID>, SQU<ARE>, RAMP, PULS<E>, NOIS<E>, DC, USER. """,
        validator=string_validator,
        values=["SINUSOID", "SIN", "SQUARE", "SQU", "RAMP",
                "PULSE", "PULS", "NOISE", "NOIS", "DC", "USER"],
    )

    frequency = Instrument.control(
        "FREQ?", "FREQ %s",
        """ A floating point property that controls the frequency of the output
        waveform in Hz, from 1e-6 (1 uHz) to 20e+6 (20 MHz), depending on the
        specified function. Can be set. """,
        validator=strict_range,
        values=[1e-6, 5e+6],
    )

    amplitude = Instrument.control(
        "VOLT?", "VOLT %f",
        """ A floating point property that controls the voltage amplitude of the
        output waveform in V, from 10e-3 V to 10 V. Can be set. """,
        validator=strict_range,
        values=[10e-3, 10],
    )

    amplitude_unit = Instrument.control(
        "VOLT:UNIT?", "VOLT:UNIT %s",
        """ A string property that controls the units of the amplitude. Valid
        values are Vpp (default), Vrms, and dBm. Can be set. """,
        validator=string_validator,
        values=["VPP", "VRMS", "DBM"],
    )

    offset = Instrument.control(
        "VOLT:OFFS?", "VOLT:OFFS %f",
        """ A floating point property that controls the voltage offset of the
        output waveform in V, from 0 V to 4.995 V, depending on the set
        voltage amplitude (maximum offset = (10 - voltage) / 2). Can be set.
        """,
        validator=strict_range,
        values=[-4.995, +4.995],
    )

    voltage_high = Instrument.control(
        "VOLT:HIGH?", "VOLT:HIGH %f",
        """ A floating point property that controls the upper voltage of the
        output waveform in V, from -4.990 V to 5 V (must be higher than low
        voltage). Can be set. """,
        validator=strict_range,
        values=[-4.99, 5],
    )

    voltage_low = Instrument.control(
        "VOLT:LOW?", "VOLT:LOW %f",
        """ A floating point property that controls the lower voltage of the
        output waveform in V, from -5 V to 4.990 V (must be lower than high
        voltage). Can be set. """,
        validator=strict_range,
        values=[-5, 4.99],
    )

    square_dutycycle = Instrument.control(
        "FUNC:SQU:DCYC?", "FUNC:SQU:DCYC %f",
        """ A floating point property that controls the duty cycle of a square
        waveform function in percent. Can be set. """,
        validator=strict_range,
        values=[20, 80],
    )

    ramp_symmetry = Instrument.control(
        "FUNC:RAMP:SYMM?", "FUNC:RAMP:SYMM %f",
        """ A floating point property that controls the symmetry percentage
        for the ramp waveform. Can be set. """,
        validator=strict_range,
        values=[0, 100],
    )

    pulse_period = Instrument.control(
        "PULS:PER?", "PULS:PER %f",
        """ A floating point property that controls the period of a pulse
        waveform function in seconds, ranging from 200 ns to 2000 s. Can be set
        and overwrites the frequency for *all* waveforms. If the period is
        shorter than the pulse width + the edge time, the edge time and pulse
        width will be adjusted accordingly. """,
        validator=strict_range,
        values=[200e-9, 2e3],
    )

    pulse_hold = Instrument.control(
        "FUNC:PULS:HOLD?", "FUNC:PULS:HOLD %s",
        """ A string property that controls if either the pulse width or the
        duty cycle is retained when changing the period or frequency of the
        waveform. Can be set to: WIDT<H> or DCYC<LE>. """,
        validator=string_validator,
        values=["WIDT", "WIDTH", "DCYC", "DCYCLE"],
    )

    pulse_width = Instrument.control(
        "FUNC:PULS:WIDT?", "FUNC:PULS:WIDT %f",
        """ A floating point property that controls the width of a pulse
        waveform function in seconds, ranging from 20 ns to 2000 s, within a
        set of restrictions depending on the period. Can be set. """,
        validator=strict_range,
        values=[20e-9, 2e3],
    )

    pulse_dutycycle = Instrument.control(
        "FUNC:PULS:DCYC?", "FUNC:PULS:DCYC %f",
        """ A floating point property that controls the duty cycle of a pulse
        waveform function in percent. Can be set. """,
        validator=strict_range,
        values=[0, 100],
    )

    pulse_transition = Instrument.control(
        "FUNC:PULS:TRAN?", "FUNC:PULS:TRAN %f",
        """ A floating point property that controls the the edge time in
        seconds for both the rising and falling edges. It is defined as the
        time between 0.1 and 0.9 of the threshold. Valid values are between
        5 ns to 100 ns. Can be set. """,
        validator=strict_range,
        values=[5e-9, 100e-9],
    )

    output = Instrument.control(
        "OUTP?", "OUTP %d",
        """ A boolean property that turns on (True) or off (False) the output
        of the function generator. Can be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, False: 0},
    )

    burst_state = Instrument.control(
        "BURS:STAT?", "BURS:STAT %d",
        """ A boolean property that controls whether the burst mode is on
        (True) or off (False). Can be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, False: 0},
    )

    burst_mode = Instrument.control(
        "BURS:MODE?", "BURS:MODE %s",
        """ A string property that controls the burst mode. Valid values
        are: TRIG<GERED>, GAT<ED>. This setting can be set. """,
        validator=string_validator,
        values=["TRIG", "TRIGGERED", "GAT", "GATED"],
    )

    burst_ncycles = Instrument.control(
        "BURS:NCYC?", "BURS:NCYC %d",
        """ An integer property that sets the number of cycles to be output
        when a burst is triggered. Valid values are 1 to 50000. This can be
        set. """,
        validator=strict_discrete_set,
        values=range(1, 50001),
    )

    def trigger(self):
        """ Send a trigger signal to the function generator. """
        self.write("*TRG;*WAI")

    def wait_for_trigger(self, timeout=3600, should_stop=lambda: False):
        """ Wait until the triggering has finished or timeout is reached.

        :param timeout: The maximum time the waiting is allowed to take. If
                        timeout is exceeded, a TimeoutError is raised. If
                        timeout is set to zero, no timeout will be used.
        :param should_stop: Optional function (returning a bool) to allow the
                            waiting to be stopped before its end.

        """
        self.write("*OPC?")

        t0 = time()
        while True:
            try:
                ready = bool(self.read())
            except VisaIOError:
                ready = False

            if ready:
                return

            if timeout != 0 and time() - t0 > timeout:
                raise TimeoutError(
                    "Timeout expired while waiting for the Agilent 33220A" +
                    " to finish the triggering."
                )

            if should_stop:
                return

    trigger_source = Instrument.control(
        "TRIG:SOUR?", "TRIG:SOUR %s",
        """ A string property that controls the trigger source. Valid values
        are: IMM<EDIATE> (internal), EXT<ERNAL> (rear input), BUS (via trigger
        command). This setting can be set. """,
        validator=string_validator,
        values=["IMM", "IMMEDIATE", "EXT", "EXTERNAL", "BUS"],
    )

    trigger_state = Instrument.control(
        "OUTP:TRIG?", "OUTP:TRIG %d",
        """ A boolean property that controls whether the output is triggered
        (True) or not (False). Can be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, False: 0},
    )

    remote_local_state = Instrument.setting(
        "SYST:COMM:RLST %s",
        """ A string property that controls the remote/local state of the
        function generator. Valid values are: LOC<AL>, REM<OTE>, RWL<OCK>.
        This setting can only be set. """,
        validator=string_validator,
        values=["LOC", "LOCAL", "REM", "REMOTE", "RWL", "RWLOCK"],
    )

    def check_errors(self):
        """ Read all errors from the instrument. """

        errors = []
        while True:
            err = self.values("SYST:ERR?")
            if int(err[0]) != 0:
                errmsg = "Agilent 33220A: %s: %s" % (err[0], err[1])
                log.error(errmsg + '\n')
                errors.append(errmsg)
            else:
                break

        return errors

    beeper_state = Instrument.control(
        "SYST:BEEP:STAT?", "SYST:BEEP:STAT %d",
        """ A boolean property that controls the state of the beeper. Can
        be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, False: 0},
    )

    def beep(self):
        """ Causes a system beep. """
        self.write("SYST:BEEP")



#
# This file is part of the PyMeasure package.
#
# Copyright (c) 2013-2021 PyMeasure Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

# Parts of this code were copied and adapted from the Agilent33220A class.

import logging
from pymeasure.instruments import Instrument
from pymeasure.instruments.validators import strict_discrete_set,\
    strict_range
from time import time
from pyvisa.errors import VisaIOError


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


# Capitalize string arguments to allow for better conformity with other WFG's
# FIXME: Currently not used since it does not combine well with the strict_discrete_set validator
# def capitalize_string(string: str, *args, **kwargs):
#     return string.upper()


# Combine the capitalize function and validator
# FIXME: This validator is not doing anything other then self.capitalize_string
# FIXME: I removed it from this class for now
# string_validator = joined_validators(capitalize_string, strict_discrete_set)


class Agilent33500(Instrument):
    """Represents the Agilent 33500 Function/Arbitrary Waveform Generator family.
    Individual devices are represented by subclasses.

    .. code-block:: python

        generator = Agilent33500("GPIB::1")

        generator.shape = 'SIN'                 # Sets the output signal shape to sine
        generator.frequency = 1e3               # Sets the output frequency to 1 kHz
        generator.amplitude = 5                 # Sets the output amplitude to 5 Vpp
        generator.output = 'on'                 # Enables the output

        generator.shape = 'ARB'                 # Set shape to arbitrary
        generator.arb_srate = 1e6               # Set sample rate to 1MSa/s

        generator.data_volatile_clear()         # Clear volatile internal memory
        generator.data_arb(                     # Send data points of arbitrary waveform
            'test',
            range(-10000, 10000, +20),          # In this case a simple ramp
            data_format='DAC'                   # Data format is set to 'DAC'
        )
        generator.arb_file = 'test'             # Select the transmitted waveform 'test'

    """

    id = Instrument.measurement(
        "*IDN?", """ Reads the instrument identification """
    )

    def __init__(self, adapter, **kwargs):
        super(Agilent33500, self).__init__(
            adapter,
            "Agilent 33500 Function/Arbitrary Waveform generator family",
            **kwargs
        )

    def beep(self):
        """ Causes a system beep. """
        self.write("SYST:BEEP")

    shape = Instrument.control(
        "FUNC?", "FUNC %s",
        """ A string property that controls the output waveform. Can be set to:
        SIN<USOID>, SQU<ARE>, TRI<ANGLE>, RAMP, PULS<E>, PRBS,  NOIS<E>, ARB, DC. """,
        validator=strict_discrete_set,
        values=["SIN", "SQU", "TRI", "RAMP", "PULS", "PRBS", "NOIS", "ARB", "DC"],
    )

    frequency = Instrument.control(
        "FREQ?", "FREQ %f",
        """ A floating point property that controls the frequency of the output
        waveform in Hz, from 1 uHz to 120 MHz (maximum range, can be lower depending
        on your device), depending on the specified function. Can be set. """,
        validator=strict_range,
        values=[1e-6, 120e+6],
    )

    amplitude = Instrument.control(
        "VOLT?", "VOLT %f",
        """ A floating point property that controls the voltage amplitude of the
        output waveform in V, from 10e-3 V to 10 V. Depends on the output
        impedance. Can be set. """,
        validator=strict_range,
        values=[10e-3, 10],
    )

    amplitude_unit = Instrument.control(
        "VOLT:UNIT?", "VOLT:UNIT %s",
        """ A string property that controls the units of the amplitude. Valid
        values are VPP (default), VRMS, and DBM. Can be set. """,
        validator=strict_discrete_set,
        values=["VPP", "VRMS", "DBM"],
    )

    offset = Instrument.control(
        "VOLT:OFFS?", "VOLT:OFFS %f",
        """ A floating point property that controls the voltage offset of the
        output waveform in V, from 0 V to 4.995 V, depending on the set
        voltage amplitude (maximum offset = (Vmax - voltage) / 2). Can be set.
        """,
        validator=strict_range,
        values=[-4.995, +4.995],
    )

    voltage_high = Instrument.control(
        "VOLT:HIGH?", "VOLT:HIGH %f",
        """ A floating point property that controls the upper voltage of the
        output waveform in V, from -4.990 V to 5 V (must be higher than low
        voltage by at least 1 mV). Can be set. """,
        validator=strict_range,
        values=[-4.99, 5],
    )

    voltage_low = Instrument.control(
        "VOLT:LOW?", "VOLT:LOW %f",
        """ A floating point property that controls the lower voltage of the
        output waveform in V, from -5 V to 4.990 V (must be lower than high
        voltage by at least 1 mV). Can be set. """,
        validator=strict_range,
        values=[-5, 4.99],
    )

    square_dutycycle = Instrument.control(
        "FUNC:SQU:DCYC?", "FUNC:SQU:DCYC %f",
        """ A floating point property that controls the duty cycle of a square
        waveform function in percent, from 0.01% to 99.98%.
        The duty cycle is limited by the frequency and the minimal pulse width of
        16 ns. See manual for more details. Can be set. """,
        validator=strict_range,
        values=[0.01, 99.98],
    )

    ramp_symmetry = Instrument.control(
        "FUNC:RAMP:SYMM?", "FUNC:RAMP:SYMM %f",
        """ A floating point property that controls the symmetry percentage
        for the ramp waveform, from 0.0% to 100.0% Can be set. """,
        validator=strict_range,
        values=[0, 100],
    )

    pulse_period = Instrument.control(
        "FUNC:PULS:PER?", "FUNC:PULS:PER %e",
        """ A floating point property that controls the period of a pulse
        waveform function in seconds, ranging from 33 ns to 1e6 s. Can be set
        and overwrites the frequency for *all* waveforms. If the period is
        shorter than the pulse width + the edge time, the edge time and pulse
        width will be adjusted accordingly. """,
        validator=strict_range,
        values=[33e-9, 1e6],
    )

    pulse_hold = Instrument.control(
        "FUNC:PULS:HOLD?", "FUNC:PULS:HOLD %s",
        """ A string property that controls if either the pulse width or the
        duty cycle is retained when changing the period or frequency of the
        waveform. Can be set to: WIDT<H> or DCYC<LE>. """,
        validator=strict_discrete_set,
        values=["WIDT", "WIDTH", "DCYC", "DCYCLE"],
    )

    pulse_width = Instrument.control(
        "FUNC:PULS:WIDT?", "FUNC:PULS:WIDT %e",
        """ A floating point property that controls the width of a pulse
        waveform function in seconds, ranging from 16 ns to 1e6 s, within a
        set of restrictions depending on the period. Can be set. """,
        validator=strict_range,
        values=[16e-9, 1e6],
    )

    pulse_dutycycle = Instrument.control(
        "FUNC:PULS:DCYC?", "FUNC:PULS:DCYC %f",
        """ A floating point property that controls the duty cycle of a pulse
        waveform function in percent, from 0% to 100%. Can be set. """,
        validator=strict_range,
        values=[0, 100],
    )

    pulse_transition = Instrument.control(
        "FUNC:PULS:TRAN?", "FUNC:PULS:TRAN:BOTH %e",
        """ A floating point property that controls the edge time in
        seconds for both the rising and falling edges. It is defined as the
        time between the 10% and 90% thresholds of the edge.
        Valid values are between 8.4 ns to 1 µs. Can be set. """,
        validator=strict_range,
        values=[8.4e-9, 1e-6],
    )

    output = Instrument.control(
        "OUTP?", "OUTP %d",
        """ A boolean property that turns on (True, 'on') or off (False, 'off') 
        the output of the function generator. Can be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, 'on': 1, 'ON': 1, False: 0, 'off': 0, 'OFF': 0},
    )

    output_load = Instrument.control(
        "OUTP:LOAD?", "OUTP:LOAD %s",
        """ Sets the expected load resistance (should be the load impedance connected
        to the output. The output impedance is always 50 Ohm, this setting can be used
        to correct the displayed voltage for loads unmatched to 50 Ohm.
        Valid values are between 1 and 10 kOhm or INF for high impedance.
        No validator is used since both numeric and string inputs are accepted,
        thus a value outside the range will not return an error.
        Can be set. """,
    )

    burst_state = Instrument.control(
        "BURS:STAT?", "BURS:STAT %d",
        """ A boolean property that controls whether the burst mode is on
        (True) or off (False). Can be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, False: 0},
    )

    burst_mode = Instrument.control(
        "BURS:MODE?", "BURS:MODE %s",
        """ A string property that controls the burst mode. Valid values
        are: TRIG<GERED>, GAT<ED>. This setting can be set. """,
        validator=strict_discrete_set,
        values=["TRIG", "TRIGGERED", "GAT", "GATED"],
    )

    burst_period = Instrument.control(
        "BURS:INT:PER?", "BURS:INT:PER %e",
        """ A floating point property that controls the period of subsequent bursts.
        Has to follow the equation burst_period > (burst_ncycles / frequency) + 1 µs.
        Valid values are 1 µs to 8000 s. Can be set. """,
        validator=strict_range,
        values=[1e-6, 8000],
    )

    burst_ncycles = Instrument.control(
        "BURS:NCYC?", "BURS:NCYC %d",
        """ An integer property that sets the number of cycles to be output
        when a burst is triggered. Valid values are 1 to 100000. This can be
        set. """,
        validator=strict_range,
        values=range(1, 100000),
    )

    arb_file = Instrument.control(
        "FUNC:ARB?", "FUNC:ARB %s",
        """ A string property that selects the arbitrary signal from the volatile
        memory of the device. String has to match an existing arb signal in volatile
        memore (set by data_arb()). Can be set. """
    )

    arb_advance = Instrument.control(
        "FUNC:ARB:ADV?", "FUNC:ARB:ADV %s",
        """ A string property that selects how the device advances from data point
        to data point. Can be set to 'TRIG<GER>' or 'SRAT<E>' (default). """,
        validator=strict_discrete_set,
        values=["TRIG", "TRIGGER", "SRAT", "SRATE"],
    )

    arb_filter = Instrument.control(
        "FUNC:ARB:FILT?", "FUNC:ARB:FILT %s",
        """ A string property that selects the filter setting for arbitrary signals. 
        Can be set to 'NORM<AL>', 'STEP' and 'OFF'. """,
        validator=strict_discrete_set,
        values=["NORM", "NORMAL", "STEP", "OFF"],
    )

    # TODO: This implementation is currently not working. Do not know why.
    # arb_period = Instrument.control(
    #     "FUNC:ARB:PER?", "FUNC:ARB:PER %e",
    #     """ A floating point property that controls the period of the arbitrary signal.
    #     Limited by number of signal points. Check for instrument errors when setting
    #     this property. Can be set. """,
    #     validator=strict_range,
    #     values=[33e-9, 1e6],
    # )
    #
    # arb_frequency = Instrument.control(
    #     "FUNC:ARB:FREQ?", "FUNC:ARB:FREQ %f",
    #     """ A floating point property that controls the frequency of the arbitrary signal.
    #     Limited by number of signal points. Check for instrument
    #     errors when setting this property. Can be set. """,
    #     validator=strict_range,
    #     values=[1e-6, 30e+6],
    # )
    #
    # arb_npoints = Instrument.measurement(
    #     "FUNC:ARB:POIN?",
    #     """ Returns the number of points in the currently selected arbitrary trace. """
    # )
    #
    # arb_voltage = Instrument.control(
    #     "FUNC:ARB:PTP?", "FUNC:ARB:PTP %f",
    #     """ An floating point property that sets the peak-to-peak voltage for the
    #     currently selected arbitrary signal. Valid values are 1 mV to 10 V. This can be
    #     set. """,
    #     validator=strict_range,
    #     values=[0.001, 10],
    # )

    arb_srate = Instrument.control(
        "FUNC:ARB:SRAT?", "FUNC:ARB:SRAT %f",
        """ An floating point property that sets the sample rate of the currently selected 
        arbitrary signal. Valid values are 1 µSa/s to 250 MSa/s (maximum range, can be lower
        depending on your device). This can be set. """,
        validator=strict_range,
        values=[1e-6, 250e6],
    )

    def data_volatile_clear(self):
        """
        Clear all arbitrary signals from the volatile memory. This should be done if the same name
        is used continuously to load different arbitrary signals into the memory, since an error
        will occur if a trace is loaded which already exists in the memory.
        """
        self.write("DATA:VOL:CLE")

    def data_arb(self, arb_name, data_points, data_format='DAC'):
        """
        Uploads an arbitrary trace into the volatile memory of the device. The data_points can be given
        as comma separated 16 bit DAC values (ranging from -32767 to +32767), as comma separated floating
        point values (ranging from -1.0 to +1.0) or as a binary data stream. Check the manual for more
        information. The storage depends on the device type and ranges from 8 Sa to 16 MSa (maximum).
        TODO: *Binary is not yet implemented*

        :param arb_name: The name of the trace in the volatile memory. This is used to access the
                         trace.
        :param data_points: Individual points of the trace. The format depends on the format
                            parameter.

                            format = 'DAC' (default): Accepts list of integer values ranging from
                            -32767 to +32767. Minimum of 8 a maximum of 65536 points.

                            format = 'float': Accepts list of floating point values ranging from
                            -1.0 to +1.0. Minimum of 8 a maximum of 65536 points.

                            format = 'binary': Accepts a binary stream of 8 bit data.
        :param data_format: Defines the format of data_points. Can be 'DAC' (default), 'float' or
                            'binary'. See documentation on parameter data_points above.
        """
        if data_format == 'DAC':
            separator = ', '
            data_points_str = [str(item) for item in data_points]  # Turn list entries into strings
            data_string = separator.join(data_points_str)  # Join strings with separator
            print("DATA:ARB:DAC {}, {}".format(arb_name, data_string))
            self.write("DATA:ARB:DAC {}, {}".format(arb_name, data_string))
            return
        elif data_format == 'float':
            separator = ', '
            data_points_str = [str(item) for item in data_points]  # Turn list entries into strings
            data_string = separator.join(data_points_str)  # Join strings with separator
            print("DATA:ARB {}, {}".format(arb_name, data_string))
            self.write("DATA:ARB {}, {}".format(arb_name, data_string))
            return
        elif data_format == 'binary':
            raise NotImplementedError('The binary format has not yet been implemented. Use "DAC" or "float" instead.')
        else:
            raise ValueError('Undefined format keyword was used. Valid entries are "DAC", "float" and "binary"')

    display = Instrument.setting(
        "DISP:TEXT '%s'",
        """ A string property which is displayed on the front panel of
        the device. Can be set. """,
    )

    def clear_display(self):
        """ Removes a text message from the display. """
        self.write("DISP:TEXT:CLE")

    def trigger(self):
        """ Send a trigger signal to the function generator. """
        self.write("*TRG;*WAI")

    def wait_for_trigger(self, timeout=3600, should_stop=lambda: False):
        """ Wait until the triggering has finished or timeout is reached.

        :param timeout: The maximum time the waiting is allowed to take. If
                        timeout is exceeded, a TimeoutError is raised. If
                        timeout is set to zero, no timeout will be used.
        :param should_stop: Optional function (returning a bool) to allow the
                            waiting to be stopped before its end.

        """
        self.write("*OPC?")

        t0 = time()
        while True:
            try:
                ready = bool(self.read())
            except VisaIOError:
                ready = False

            if ready:
                return

            if timeout != 0 and time() - t0 > timeout:
                raise TimeoutError(
                    "Timeout expired while waiting for the Agilent 33220A" +
                    " to finish the triggering."
                )

            if should_stop:
                return

    trigger_source = Instrument.control(
        "TRIG:SOUR?", "TRIG:SOUR %s",
        """ A string property that controls the trigger source. Valid values
        are: IMM<EDIATE> (internal), EXT<ERNAL> (rear input), BUS (via trigger
        command). This setting can be set. """,
        validator=strict_discrete_set,
        values=["IMM", "IMMEDIATE", "EXT", "EXTERNAL", "BUS"],
    )

    ext_trig_out = Instrument.control(
        "OUTP:TRIG?", "OUTP:TRIG %d",
        """ A boolean property that controls whether the trigger out signal is
        active (True) or not (False). This signal is output from the Ext Trig
        connector on the rear panel in Burst and Wobbel mode. Can be set. """,
        validator=strict_discrete_set,
        map_values=True,
        values={True: 1, False: 0},
    )

    def check_errors(self):
        """ Read all errors from the instrument. """

        errors = []
        while True:
            err = self.values("SYST:ERR?")
            if int(err[0]) != 0:
                errmsg = "Agilent 33521A: %s: %s" % (err[0], err[1])
                log.error(errmsg + '\n')
                errors.append(errmsg)
            else:
                break

        return errors
